from visionsuite.engines.utils.bases.base_results import BaseResults, update_log
from visionsuite.engines.classification.utils.results.results import Results


class ValResults(Results):
    __instance = None
    __initialized = False

    __best_accuracy = -1

    @classmethod
    def get_attribute_names(cls):
        attribute_names = list(cls.__dict__.keys())
        return attribute_names

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            cls._log = {}

        return cls.__instance

    def __init__(self):
        if self.__initialized:
            return

        super().__init__()
        self.__initialized = True

    def __setattr__(self, name, value):
        allowed_attributes = (
            self.get_attribute_names()
            + Results.get_attribute_names()
            + BaseResults.get_attribute_names()
        )
        if name not in allowed_attributes:
            raise AttributeError(f"Attribute '{name}' is not allowed in this class")
        super().__setattr__(name, value)

    @property
    def best_loss(self):
        return self.__best_loss

    @best_loss.setter
    @update_log
    def best_loss(self, value):
        self.__best_loss = value

    @property
    def best_accuracy(self):
        return self.__best_accuracy

    @best_accuracy.setter
    @update_log
    def best_accuracy(self, value):
        self.__best_accuracy = value


if __name__ == "__main__":
    val_res1 = ValResults()
    val_res2 = ValResults()

    loss = 0.1
    cpu_usage = 8.5
    gpu_usage = 8.1
    iou = 0.1
    time_for_a_epoch = 2
    time_for_epochs = 2
    bg_iou = 0.1
    stabbed_iou = 0.05

    for epoch in range(1, 5):
        loss -= 0.02
        cpu_usage += 0.01
        gpu_usage += 0.02
        iou += 0.05
        time_for_a_epoch += 0.01
        time_for_epochs += 0.05
        bg_iou += 0.1
        stabbed_iou += 0.05

        val_res1.epoch = epoch
        val_res1.loss = loss
        val_res1.cpu_usage = cpu_usage
        val_res1.gpu_usage = gpu_usage
        val_res1.iou = iou
        val_res1.time_for_a_epoch = time_for_a_epoch
        val_res1.time_for_epochs = time_for_epochs
        val_res1.classes_iou = {"bg_iou": bg_iou, "stabbed_iou": stabbed_iou}

        print("1. ", val_res1.log)
        print("2. ", val_res2.log)



### TrainRunner
def on_train_runner_set_configs(runner, *args, **kwargs):
    pass

def on_train_runner_set_variables(runner, *args, **kwargs):
    if runner.args['augment']['train']['backend'].lower() != "pil" and not runner.args['augment']['train']['use_v2']:
        # TODO: Support tensor backend in V1?
        raise ValueError("Use --use-v2 if you want to use the tv_tensor or tensor backend.")
    if runner.args['augment']['train']['use_v2'] and runner.args['dataset']['type'] != "coco":
        raise ValueError("v2 is only support supported for coco dataset for now.")

def on_train_runner_run_start(runner, *args, **kwargs):
    pass

def on_train_runner_run_end(runner, *args, **kwargs):
    pass

train_callbacks = {
    "on_runner_set_configs": [on_train_runner_set_configs],
    "on_runner_set_variables": [on_train_runner_set_variables],
    "on_runner_run_start": [on_train_runner_run_start],
    "on_runner_run_end": [on_train_runner_run_end],
}


### TestRunner
def on_test_runner_set_configs(runner, *args, **kwargs):
    pass

def on_test_runner_set_variables(runner, *args, **kwargs):
    if runner.args['augment']['test']['backend'].lower() != "pil" and not runner.args['augment']['test']['use_v2']:
        # TODO: Support tensor backend in V1?
        raise ValueError("Use --use-v2 if you want to use the tv_tensor or tensor backend.")
    if runner.args['augment']['test']['use_v2'] and runner.args['dataset']['type'] != "coco":
        raise ValueError("v2 is only support supported for coco dataset for now.")

def on_test_runner_run_start(runner, *args, **kwargs):
    pass

def on_test_runner_run_end(runner, *args, **kwargs):
    pass

test_callbacks = {
    "on_runner_set_configs": [on_train_runner_set_configs],
    "on_runner_set_variables": [on_train_runner_set_variables],
    "on_runner_run_start": [on_train_runner_run_start],
    "on_runner_run_end": [on_train_runner_run_end],
}


import datetime
import json
import logging
import logging.config
import os
import os.path as osp
import sys

# import socket
import warnings
from pathlib import Path

ROOT = Path(__file__).resolve().parents[0]
LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class Logger:
    """
    arguments:
        - log_dir: directory to save log file
        - log_stream_level: one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', and 'CRITICAL'
        - log_file_level: one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', and 'CRITICAL'
        - config_json: default is already set
    """

    def __init__(self, name=None):
        if name is None:
            self.__name = "no_name"
        else:
            self.__name = name
        self._logger = None
        self.is_logger_set = False

        self.__log_dir = None
        self.__log_stream_level = "DEBUG"
        self.__log_file_level = "DEBUG"
        self.__config_json = None
        self.__log_filename = None

        self.__config = None

    def __repr__(self):
        if self._logger is None:
            return f"The logger({self._logger}) is not yet defined. Need to execute 'set_logger' first"
        else:
            return f"This logger is for {self.name} and locates in {self.log_dir if self.log_filename is None else self.log_filename} with file-level({self.__log_file_level}) and stream-level({self.__log_stream_level})"

    @property
    def log_dir(self):
        return self.__log_dir

    @log_dir.setter
    def log_dir(self, value):
        self.__log_dir = value

    @property
    def log_filename(self):
        return self.__log_filename

    @log_filename.setter
    def log_filename(self, value):
        self.__log_filename = value

    @property
    def log_stream_level(self):
        return self.__log_stream_level

    @log_stream_level.setter
    def log_stream_level(self, value):
        assert value.upper() in self._get_log_levels(), ValueError(
            f"Log-level for streaming should be one of {LOG_LEVELS}, not {value}"
        )
        self.__log_stream_level = value.upper()

    @property
    def log_file_level(self):
        return self.__log_file_level

    @log_file_level.setter
    def log_file_level(self, value):
        assert value.upper() in self._get_log_levels(), ValueError(
            f"Log-level for file should be one of {LOG_LEVELS}, not {value}"
        )
        self.__log_file_level = value.upper()

    @property
    def config_json(self):
        return self.__config_json

    @config_json.setter
    def config_json(self, value):
        assert osp.exists(value), ValueError(f"There is no such config_json({value})")

        self.__config_json = value

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, value):
        self.__config = value

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    def set_logger(
        self,
        log_stream_level="DEBUG",
        log_file_level="DEBUG",
        log_dir=None,
        config_json=osp.join(ROOT, "data/colored-logging.json"),
    ):
        if self.is_logger_set:
            self.log_info(f"Logger is already set for {self._name}")
            return 
        self.log_stream_level = log_stream_level.upper()
        self.log_file_level = log_file_level.upper()
        self.log_dir = log_dir
        self.config_json = config_json

        # FIXME: logging.config.dictlogging.config.dictConfigConfig가 오버랩되어서 사용되어 다중 logger가 불가능,,, json을 변경해야한다고 함
        # try:
        #     self._get_config()
        #     logging.config.dictConfig(self.config)
        #     self._logger = logging.getLogger(self.name)
        #     self.is_logger_set = True
        # except Exception as e:
        #     warnings.warn(f"Cannot define logger: {e}")
        #     try:
        #         self.configs_json = osp.join(ROOT, 'data/logging.json')
        #         self._get_config()
        #         logging.config.dictConfig(self.config)
        #         self._logger = logging.getLogger(self.name)
        #         self.is_logger_set = True
        #     except Exception as e:
        #         warnings.warn(f"Cannot define logger: {e}")
        #         self._logger = None
        #         self.is_logger_set = False

        # self.log_info(f"Logger has been set for {self.__name}", self.__init__.__name__, __class__.__name__)

        try:
            self._get_config()
            self._logger = logging.getLogger(self.name)
            self._logger.setLevel(self.log_stream_level)
            file_handler = logging.FileHandler(
                osp.join(self.__log_dir, self.__name + ".log")
            )
            file_handler.setLevel(self.log_file_level)
            file_handler.setFormatter(
                logging.Formatter(self.config["formatters"]["colored"]["format"])
            )
            self._logger.addHandler(file_handler)

            console_handler = logging.StreamHandler()
            console_handler.setLevel(self.log_stream_level)
            console_handler.setFormatter(
                logging.Formatter(self.config["formatters"]["colored"]["format"])
            )
            self._logger.addHandler(console_handler)

            self.is_logger_set = True
        except Exception as e:
            warnings.warn(f"Cannot define logger: {e}")
            try:
                self.configs_json = osp.join(ROOT, "data/logging.json")
                self._get_config()
                self._logger = logging.getLogger(self.name)
                file_handler = logging.FileHandler(
                    osp.join(self.__log_dir, self.__name + ".log")
                )
                file_handler.setFormatter(
                    logging.Formatter(self.config["formatters"]["colored"]["format"])
                )
                self._logger.addHandler(file_handler)

                console_handler = logging.StreamHandler()
                console_handler.setFormatter(
                    logging.Formatter(self.config["formatters"]["colored"]["format"])
                )
                self._logger.addHandler(console_handler)
                self.is_logger_set = True
            except Exception as e:
                warnings.warn(f"Cannot define logger: {e}")
                self._logger = None
                self.is_logger_set = False

        self.log_info(
            f"Logger has been set for {self.__name}",
            self.__init__.__name__,
            __class__.__name__,
        )

    def log_debug(self, msg, func_name="", class_name=""):
        if self._logger is not None:
            self._logger.debug(f"[{class_name}] - [{func_name}] - {msg}")

    def log_info(self, msg, func_name="", class_name=""):
        if self._logger is not None:
            self._logger.info(f"[{class_name}] - [{func_name}] - {msg}")

    def log_warning(self, msg, func_name="", class_name=""):
        if self._logger is not None:
            self._logger.warning(f"[{class_name}] - [{func_name}] - {msg}")

    def log_error(self, msg, func_name="", class_name=""):
        if self._logger is not None:
            self._logger.error(f"[{class_name}] - [{func_name}] - {msg}")

    def log_critical(self, msg, func_name="", class_name=""):
        if self._logger is not None:
            self._logger.critical(f"[{class_name}] - [{func_name}] - {msg}")

    def _get_config(self):
        with open(self.__config_json) as f:
            self.__config = json.load(f)

        self._set_log_level()
        self._set_log_dir()

    def _set_log_level(self):
        self.__config["handlers"]["stream"]["level"] = self.__log_stream_level
        self.__config["root"]["level"] = self.__log_file_level

    def _set_log_dir(self):
        if self.__log_dir is None:
            # hostname = socket.gethostname()

            # log_dir = f'/home/{hostname}/logs'
            self.__log_dir = f"/DeepLearning/_logs/logger"
            current_date = datetime.datetime.now()
            year = current_date.year
            month = current_date.month
            day = current_date.day
            hour = current_date.hour

            self.__log_dir = osp.join(
                self.__log_dir, f"{year}_{month}_{day}", str(hour)
            )

        if not osp.exists(self.__log_dir):
            os.makedirs(self.__log_dir)

        assert self.__log_dir is not None, ValueError(
            f"Log directory should be assigned, not {self.__log_dir}"
        )

        if self.__name is None:
            self.log_filename = osp.join(
                self.__log_dir, self.__config["handlers"]["file"]["filename"]
            )
        else:
            self.log_filename = osp.join(self.__log_dir, self.__name + ".log")
        self.__config["handlers"]["file"]["filename"] = self.log_filename

    def _get_log_levels(self):
        return LOG_LEVELS

    # TODO: Need to extract the argument for exit value
    def try_except_log(
        self,
        func,
        msg="",
        post_action=None,
        exit=False,
        parent_class=None,
        parent_fn=None,
    ):
        try:
            if msg == "In the post-action, ":
                if self._logger:
                    self._logger.info("Post-action runs after raising error")
            func()
        except Exception as error_msg:
            error_type = type(error_msg).__name__
            if len(msg) != 0:
                error_msg = msg + ", and " + str(error_msg)
            else:
                error_msg = str(error_msg)
            if self._logger:
                self._logger.error(
                    f" [{self.try_except_log.__name__}] {error_type}: {error_msg} at [{parent_fn}] of [{parent_class}]"
                )

            if post_action is not None:
                self.try_except_log(post_action, msg="In the post-action, ")
            if error_type in __builtins__:
                if not exit:
                    raise __builtins__[error_type](error_msg)
                else:
                    sys.exit()

    def assertion_log(
        self,
        condition,
        error,
        msg="",
        post_action=None,
        exit=False,
        parent_class=None,
        parent_fn=None,
    ):
        try:
            if msg == "In the post-action, ":
                if self._logger:
                    self._logger.info("Post-action runs after raising error")
            assert condition
        except AssertionError:
            error_type = type(error).__name__
            if len(msg) != 0:
                error_msg = msg + ", and " + str(error)
            else:
                error_msg = str(error)
            if self._logger:
                self._logger.error(
                    f" [{self.assertion_log.__name__}] {error_type}: {error_msg} at [{parent_fn}] of [{parent_class}]"
                )

            if post_action is not None:
                self.try_except_log(post_action, msg="In the post-action, ")
            if not exit:
                raise AssertionError(error_msg)
            else:
                sys.exit()

    def raise_error_log(self, error_obj, error_msg):
        if self._logger:
            self._logger.error(f"[{error_obj.__name__}] {error_msg}")
        if error_obj.__name__ in __builtins__:
            raise __builtins__[error_obj.__name__](error_msg)
        else:
            sys.exit()



# Copyright (c) Open-MMLab. All rights reserved.
import os.path as osp
import platform
import shutil

import torch
from torch.optim import Optimizer

import mmcv
from mmcv.runner import RUNNERS, EpochBasedRunner
from .checkpoint import save_checkpoint

try:
    import apex
except:
    print('apex is not installed')


@RUNNERS.register_module()
class EpochBasedRunnerAmp(EpochBasedRunner):
    """Epoch-based Runner with AMP support.

    This runner train models epoch by epoch.
    """

    def save_checkpoint(self,
                        out_dir,
                        filename_tmpl='epoch_{}.pth',
                        save_optimizer=True,
                        meta=None,
                        create_symlink=True):
        """Save the checkpoint.

        Args:
            out_dir (str): The directory that checkpoints are saved.
            filename_tmpl (str, optional): The checkpoint filename template,
                which contains a placeholder for the epoch number.
                Defaults to 'epoch_{}.pth'.
            save_optimizer (bool, optional): Whether to save the optimizer to
                the checkpoint. Defaults to True.
            meta (dict, optional): The meta information to be saved in the
                checkpoint. Defaults to None.
            create_symlink (bool, optional): Whether to create a symlink
                "latest.pth" to point to the latest checkpoint.
                Defaults to True.
        """
        if meta is None:
            meta = dict(epoch=self.epoch + 1, iter=self.iter)
        elif isinstance(meta, dict):
            meta.update(epoch=self.epoch + 1, iter=self.iter)
        else:
            raise TypeError(
                f'meta should be a dict or None, but got {type(meta)}')
        if self.meta is not None:
            meta.update(self.meta)

        filename = filename_tmpl.format(self.epoch + 1)
        filepath = osp.join(out_dir, filename)
        optimizer = self.optimizer if save_optimizer else None
        save_checkpoint(self.model, filepath, optimizer=optimizer, meta=meta)
        # in some environments, `os.symlink` is not supported, you may need to
        # set `create_symlink` to False
        if create_symlink:
            dst_file = osp.join(out_dir, 'latest.pth')
            # if platform.system() != 'Windows':
            #     mmcv.symlink(filename, dst_file)
            # else:
            shutil.copy(filepath, dst_file)

    def resume(self,
               checkpoint,
               resume_optimizer=True,
               map_location='default'):
        if map_location == 'default':
            if torch.cuda.is_available():
                device_id = torch.cuda.current_device()
                checkpoint = self.load_checkpoint(
                    checkpoint,
                    map_location=lambda storage, loc: storage.cuda(device_id))
            else:
                checkpoint = self.load_checkpoint(checkpoint)
        else:
            checkpoint = self.load_checkpoint(
                checkpoint, map_location=map_location)

        self._epoch = checkpoint['meta']['epoch']
        self._iter = checkpoint['meta']['iter']
        if 'optimizer' in checkpoint and resume_optimizer:
            if isinstance(self.optimizer, Optimizer):
                self.optimizer.load_state_dict(checkpoint['optimizer'])
            elif isinstance(self.optimizer, dict):
                for k in self.optimizer.keys():
                    self.optimizer[k].load_state_dict(
                        checkpoint['optimizer'][k])
            else:
                raise TypeError(
                    'Optimizer should be dict or torch.optim.Optimizer '
                    f'but got {type(self.optimizer)}')

        if 'amp' in checkpoint:
            apex.amp.load_state_dict(checkpoint['amp'])
            self.logger.info('load amp state dict')

        self.logger.info('resumed epoch %d, iter %d', self.epoch, self.iter)



import glob 
import os
import os.path as osp
import torch
import imgviz
import json
import cv2
import numpy as np
import pandas as pd
from mlearning.utils.vis.vis_seg import vis_seg
from athena.src.tasks.segmentation.frameworks.pytorch.models.torch_model_v2 import TorchModelV2


def get_mask_from_pred(pred, conf=0.5, contour_thres=50):
    mask = (pred > conf).astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    points = []
    for contour in contours:
        if len(contour) < contour_thres:
            pass
        else:
            epsilon = 0.001 * cv2.arcLength(contour, True)
            approx = cv2.approxPolyDP(contour, epsilon, True)[:, 0, :].tolist()
            points.append(approx)

    return mask, points

compare_mask = True
ml_framework = 'pytorch'
model_name='segnext'
weights = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/u_gap/outputs/SEGMENTATION/pytorch_segnext_b/train/weights/last.pth'
model = TorchModelV2(model_name=model_name, backbone='b', model_dict=None, 
                  width=768, height=512, channel=3, num_classes=4,
                  device='cuda:0', aux_loss=False, init_lr=1e-3, device_ids=[0], 
                  weights=weights)
classes = ['FILM', 'ROI', 'MARK']
idx2class = {idx: cls for idx, cls in enumerate(classes)}
_classes = ['ROI', 'FILM', 'MARK']
_idx2class = {idx: cls for idx, cls in enumerate(_classes)}
input_dir = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/u_gap/split_dataset/val'
json_dir = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/u_gap/split_dataset/val'
output_dir = f'/DeepLearning/etc/_athena_tests/benchmark/sungwoo/u_gap/reports/preds/athena_{model_name}'

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

results = {}
compare = {}
for img_file in img_files:
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file).astype(np.float32)
    if ml_framework == 'tensorflow':
        preds = model(img[np.newaxis, :])
        pred = preds[0].numpy()
    elif ml_framework == 'pytorch':
        torch_img = torch.from_numpy(np.transpose(img[np.newaxis, :], (0, 3, 1, 2))).to('cuda')
        preds = model(torch_img)
        pred = preds[0].detach().cpu().numpy()
        pred = np.transpose(pred, (1, 2, 0)).astype(np.uint8)

    idx2masks = {}
    for idx in idx2class.keys():
        mask, points = get_mask_from_pred(pred[:, :, idx + 1], conf=0.2)
    
        idx2masks[idx] = points
        
    if _classes is not None:
        new_idx2masks = {}
        for idx, _cls in enumerate(_classes):
            for jdx, cls in enumerate(classes):
                if cls == _cls:
                    new_idx2masks[idx] = idx2masks[jdx]
        
        idx2masks = new_idx2masks
        idx2class = _idx2class    
    
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1]
    results.update({filename: {'idx2masks': idx2masks, 'idx2class': idx2class, 'img_file': img_file}})
    
    if compare_mask:
        _compare = vis_seg(img_file, idx2masks, idx2class, output_dir, color_map, json_dir, compare_mask=compare_mask)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})
    else:
        vis_seg(img_file, idx2masks, idx2class, output_dir, color_map, json_dir, compare_mask=compare_mask)
    
with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
    json.dump(results, json_file, ensure_ascii=False, indent=4)

if compare_mask:
    with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
        json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
    df_compare = pd.DataFrame(compare)
        
    df_compare_pixel = df_compare.loc['diff_pixel'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_pixel.csv'))
    
    df_compare_pixel = df_compare.loc['diff_iou'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


import os
import sys

import pandas as pd
import numpy as np
from torch import nn
import torch
import torch.nn.functional as F


class ContrastiveLoss(nn.Module):
    def __init__(self, n_classes=19):
        super().__init__()
        self.n_classes = n_classes

    def forward(self, emb_k, emb_q, labels, epoch, tau=0.1):
        """
        emb_k: the feature bank with the aggregated embeddings over the iterations
        emb_q: the embeddings for the current iteration
        labels: the correspondent class labels for each sample in emb_q
        """
        if epoch:
            total_loss = torch.tensor(0.0).cuda()
            assert (
                emb_q.shape[0] == labels.shape[0]
            ), "mismatch on emb_q and labels shapes!"
            emb_k = F.normalize(emb_k, dim=-1)
            emb_q = F.normalize(emb_q, dim=1)

            for i, emb in enumerate(emb_q):
                label = labels[i]
                if not (255 in label.unique() and len(label.unique()) == 1):
                    label[label == 255] = self.n_classes
                    label_sq = torch.unique(label, return_inverse=True)[1]
                    oh_label = (F.one_hot(label_sq)).unsqueeze(-2)  # one hot labels
                    count = oh_label.view(-1, oh_label.shape[-1]).sum(
                        dim=0
                    )  # num of pixels per cl
                    pred = emb.permute(1, 2, 0).unsqueeze(-1)
                    oh_pred = (
                        pred * oh_label
                    )  # (H, W, Nc, Ncp) Ncp num classes present in the label
                    oh_pred_flatten = oh_pred.view(
                        oh_pred.shape[0] * oh_pred.shape[1],
                        oh_pred.shape[2],
                        oh_pred.shape[3],
                    )
                    res_raw = oh_pred_flatten.sum(dim=0) / count  # avg feat per class
                    res_new = (res_raw[~res_raw.isnan()]).view(
                        -1, self.n_classes
                    )  # filter out nans given by intermediate classes (present because of oh)
                    label_list = label.unique()
                    if self.n_classes in label_list:
                        label_list = label_list[:-1]
                        res_new = res_new[:-1, :]

                    # temperature-scaled cosine similarity
                    final = (res_new.cuda() @ emb_k.T.cuda()) / 0.1

                    loss = F.cross_entropy(final, label_list)
                    total_loss += loss

            return total_loss / emb_q.shape[0]

        return torch.tensor(0).cuda()


class OWLoss(nn.Module):
    def __init__(self, n_classes, hinged=False, delta=0.1):
        super().__init__()
        self.n_classes = n_classes
        self.hinged = hinged
        self.delta = delta
        self.count = torch.zeros(self.n_classes).cuda()  # count for class
        self.features = {
            i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)
        }
        # See https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
        # for implementation of Welford Alg.
        self.ex = {i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)}
        self.ex2 = {
            i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)
        }
        self.var = {
            i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)
        }

        self.criterion = torch.nn.L1Loss(reduction="none")

        self.previous_features = None
        self.previous_count = None

    @torch.no_grad()
    def cumulate(self, logits: torch.Tensor, sem_gt: torch.Tensor):
        sem_pred = torch.argmax(torch.softmax(logits, dim=1), dim=1)
        gt_labels = torch.unique(sem_gt).tolist()
        logits_permuted = logits.permute(0, 2, 3, 1)
        for label in gt_labels:
            if label == 255:
                continue
            sem_gt_current = sem_gt == label
            sem_pred_current = sem_pred == label
            tps_current = torch.logical_and(sem_gt_current, sem_pred_current)
            if tps_current.sum() == 0:
                continue
            logits_tps = logits_permuted[torch.where(tps_current == 1)]
            # max_values = logits_tps[:, label].unsqueeze(1)
            # logits_tps = logits_tps / max_values
            avg_mav = torch.mean(logits_tps, dim=0)
            n_tps = logits_tps.shape[0]
            # features is running mean for mav
            self.features[label] = (
                self.features[label] * self.count[label] + avg_mav * n_tps
            )

            self.ex[label] += (logits_tps).sum(dim=0)
            self.ex2[label] += ((logits_tps) ** 2).sum(dim=0)
            self.count[label] += n_tps
            self.features[label] /= self.count[label] + 1e-8

    def forward(
        self, logits: torch.Tensor, sem_gt: torch.Tensor, is_train: torch.bool
    ) -> torch.Tensor:
        if is_train:
            # update mav only at training time
            sem_gt = sem_gt.type(torch.uint8)
            self.cumulate(logits, sem_gt)
        if self.previous_features == None:
            return torch.tensor(0.0).cuda()
        gt_labels = torch.unique(sem_gt).tolist()

        logits_permuted = logits.permute(0, 2, 3, 1)

        acc_loss = torch.tensor(0.0).cuda()
        for label in gt_labels[:-1]:
            mav = self.previous_features[label]
            logs = logits_permuted[torch.where(sem_gt == label)]
            mav = mav.expand(logs.shape[0], -1)
            if self.previous_count[label] > 0:
                ew_l1 = self.criterion(logs, mav)
                ew_l1 = ew_l1 / (self.var[label] + 1e-8)
                if self.hinged:
                    ew_l1 = F.relu(ew_l1 - self.delta).sum(dim=1)
                acc_loss += ew_l1.mean()

        return acc_loss

    def update(self):
        self.previous_features = self.features
        self.previous_count = self.count
        for c in self.var.keys():
            self.var[c] = (self.ex2[c] - self.ex[c] ** 2 / (self.count[c] + 1e-8)) / (
                self.count[c] + 1e-8
            )

        # resetting for next epoch
        self.count = torch.zeros(self.n_classes)  # count for class
        self.features = {
            i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)
        }
        self.ex = {i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)}
        self.ex2 = {
            i: torch.zeros(self.n_classes).cuda() for i in range(self.n_classes)
        }

        return self.previous_features, self.var

    def read(self):
        mav_tensor = torch.zeros(self.n_classes, self.n_classes)
        for key in self.previous_features.keys():
            mav_tensor[key] = self.previous_features[key]
        return mav_tensor


class ObjectosphereLoss(nn.Module):
    def __init__(self, sigma=1.0):
        super().__init__()
        self.sigma = sigma

    def forward(self, logits, sem_gt):
        logits_unk = logits.permute(0, 2, 3, 1)[torch.where(sem_gt == 255)]
        logits_kn = logits.permute(0, 2, 3, 1)[torch.where(sem_gt != 255)]

        if len(logits_unk):
            loss_unk = torch.linalg.norm(logits_unk, dim=1).mean()
        else:
            loss_unk = torch.tensor(0)
        if len(logits_kn):
            loss_kn = F.relu(self.sigma - torch.linalg.norm(logits_kn, dim=1)).mean()
        else:
            loss_kn = torch.tensor(0)

        loss = 10 * loss_unk + loss_kn
        return loss


class CrossEntropyLoss2d(nn.Module):
    def __init__(self, device, weight):
        super(CrossEntropyLoss2d, self).__init__()
        self.weight = torch.tensor(weight).to(device)
        self.num_classes = len(self.weight) + 1  # +1 for void
        if self.num_classes < 2**8:
            self.dtype = torch.uint8
        else:
            self.dtype = torch.int16
        self.ce_loss = nn.CrossEntropyLoss(
            torch.from_numpy(np.array(weight)).float(),
            reduction="none",
            ignore_index=-1,
        )
        self.ce_loss.to(device)

    def forward(self, inputs, targets):
        losses = []
        number_of_pixels_per_class = torch.bincount(
            targets.flatten().type(self.dtype), minlength=self.num_classes
        )
        divisor_weighted_pixel_sum = torch.sum(
            number_of_pixels_per_class[1:] * self.weight
        )  # without void
        targets_m = targets.clone()
        if targets_m.sum() == 0:
            import ipdb;ipdb.set_trace()  # fmt: skip
        targets_m -= 1
        loss_all = self.ce_loss(inputs, targets_m.long())
        if divisor_weighted_pixel_sum > 0:
            losses.append(torch.sum(loss_all) / divisor_weighted_pixel_sum)
        else:
            losses.append(torch.tensor(0.0).cuda())

        return losses


class CrossEntropyLoss2dForValidData:
    def __init__(self, device, weight, weighted_pixel_sum):
        super(CrossEntropyLoss2dForValidData, self).__init__()
        self.ce_loss = nn.CrossEntropyLoss(
            torch.from_numpy(np.array(weight)).float(), reduction="sum", ignore_index=-1
        )
        self.ce_loss.to(device)
        self.weighted_pixel_sum = weighted_pixel_sum
        self.total_loss = 0

    def add_loss_of_batch(self, inputs, targets):
        targets_m = targets.clone()
        targets_m -= 1
        loss = self.ce_loss(inputs, targets_m.long())
        self.total_loss += loss

    def compute_whole_loss(self):
        return self.total_loss.cpu().numpy().item() / self.weighted_pixel_sum.item()

    def reset_loss(self):
        self.total_loss = 0


class CrossEntropyLoss2dForValidDataUnweighted:
    def __init__(self, device):
        super(CrossEntropyLoss2dForValidDataUnweighted, self).__init__()
        self.ce_loss = nn.CrossEntropyLoss(
            weight=None, reduction="sum", ignore_index=-1
        )
        self.ce_loss.to(device)
        self.nr_pixels = 0
        self.total_loss = 0

    def add_loss_of_batch(self, inputs, targets):
        targets_m = targets.clone()
        targets_m -= 1
        loss = self.ce_loss(inputs, targets_m.long())
        self.total_loss += loss
        self.nr_pixels += torch.sum(targets_m >= 0)  # only non void pixels

    def compute_whole_loss(self):
        return (
            self.total_loss.cpu().numpy().item() / self.nr_pixels.cpu().numpy().item()
        )

    def reset_loss(self):
        self.total_loss = 0
        self.nr_pixels = 0


def print_log(
    epoch, local_count, count_inter, dataset_size, loss, time_inter, learning_rates
):
    print_string = "Train Epoch: {:>3} [{:>4}/{:>4} ({: 5.1f}%)]".format(
        epoch, local_count, dataset_size, 100.0 * local_count / dataset_size
    )
    for i, lr in enumerate(learning_rates):
        print_string += "   lr_{}: {:>6}".format(i, round(lr, 10))
    print_string += "   Loss: {:0.6f}".format(loss.item())
    print_string += "  [{:0.2f}s every {:>4} data]".format(time_inter, count_inter)
    print(print_string, flush=True)


def save_ckpt(ckpt_dir, model, optimizer, epoch):
    state = {
        "epoch": epoch,
        "state_dict": model.state_dict(),
        "optimizer": optimizer.state_dict(),
    }
    ckpt_model_filename = "ckpt_epoch_{}.pth".format(epoch)
    path = os.path.join(ckpt_dir, ckpt_model_filename)
    torch.save(state, path)
    print("{:>2} has been successfully saved".format(path))


def save_ckpt_every_epoch(
    ckpt_dir, model, optimizer, epoch, best_miou, best_miou_epoch, mavs, stds
):
    state = {
        "epoch": epoch,
        "state_dict": model.state_dict(),
        "optimizer": optimizer.state_dict(),
        "best_miou": best_miou,
        "best_miou_epoch": best_miou_epoch,
        "mavs": mavs,
        "stds": stds,
    }
    ckpt_model_filename = "ckpt_latest.pth".format(epoch)
    path = os.path.join(ckpt_dir, ckpt_model_filename)
    torch.save(state, path)


def load_ckpt(model, optimizer, model_file, device):
    if os.path.isfile(model_file):
        print("=> loading checkpoint '{}'".format(model_file))
        if device.type == "cuda":
            checkpoint = torch.load(model_file)
        else:
            checkpoint = torch.load(
                model_file, map_location=lambda storage, loc: storage
            )

        mav_dict = checkpoint["mavs"]
        std_dict = checkpoint["stds"]

        model.load_state_dict(checkpoint["state_dict"])

        if optimizer:
            optimizer.load_state_dict(checkpoint["optimizer"])
        print(
            "=> loaded checkpoint '{}' (epoch {})".format(
                model_file, checkpoint["epoch"]
            )
        )
        epoch = checkpoint["epoch"]
        if "best_miou" in checkpoint:
            best_miou = checkpoint["best_miou"]
            print("Best mIoU:", best_miou)
        else:
            best_miou = 0

        if "best_miou_epoch" in checkpoint:
            best_miou_epoch = checkpoint["best_miou_epoch"]
            print("Best mIoU epoch:", best_miou_epoch)
        else:
            best_miou_epoch = 0
        return epoch, best_miou, best_miou_epoch, mav_dict, std_dict
    else:
        print("=> no checkpoint found at '{}'".format(model_file))
        sys.exit(1)


def get_best_checkpoint(ckpt_dir, key="mIoU_test"):
    ckpt_path = None
    log_file = os.path.join(ckpt_dir, "logs.csv")
    if os.path.exists(log_file):
        data = pd.read_csv(log_file)
        idx = data[key].idxmax()
        miou = data[key][idx]
        epoch = data.epoch[idx]
        ckpt_path = os.path.join(ckpt_dir, f"ckpt_epoch_{epoch}.pth")
    assert ckpt_path is not None, f"No trainings found at {ckpt_dir}"
    assert os.path.exists(ckpt_path), f"There is no weights file named {ckpt_path}"
    print(f"Best mIoU: {100*miou:0.2f} at epoch: {epoch}")
    return ckpt_path



import os 
import os.path as osp 
import glob 
from shutil import copyfile
import json
from visionsuite.utils.helpers import get_filename
import random
from tqdm import tqdm


def split_dataset(input_dir, output_dir, image_formats, ratio, figs=True):
    if not osp.exists(output_dir):
        os.mkdir(output_dir)

    img_files = [file for image_format in image_formats for file in glob.glob(osp.join(input_dir, f'*/*.{image_format}'))]

    label2idx = {}
    label2cnt, train_label2cnt, val_label2cnt = {}, {}, {}
    for img_file in tqdm(img_files):
        
        json_file = osp.splitext(img_file)[0] + '.json'
        with open(json_file, 'r') as jf:
            anns = json.load(jf)['shapes']
            
        labels = []
        for ann in anns:
            label = ann['label']
            if label not in label2idx:
                label2idx[label] = len(label2idx)
                label2cnt[label], train_label2cnt[label], val_label2cnt[label] = 0, 0, 0
                
            label2cnt[label] += 1
            labels.append(label)
            
        if random.uniform(0, 1) <= ratio:
            mode = 'val'
        else:
            mode = 'train'
            
        for label in labels:
            if int(train_label2cnt[label]*(ratio/(1 - ratio))) > val_label2cnt[label]:
                mode = 'val'
                break
            
        if mode == 'train':
            for label in labels:
                train_label2cnt[label] += 1
        elif mode == 'val':
            for label in labels:
                val_label2cnt[label] += 1
        

        _output_dir = osp.join(output_dir, mode)
        if not osp.exists(_output_dir):
            os.mkdir(_output_dir)
            
        copyfile(img_file, osp.join(_output_dir, get_filename(img_file, True)))            
        copyfile(json_file, osp.join(_output_dir, get_filename(json_file, True)))            
        
        
    f = open(osp.join(output_dir, 'classes.txt'), 'w')
    for key, val in label2idx.items():
        f.write(f"{val}: {key}\n")        
    f.close()

    if figs:
        import plotly.graph_objects as go

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=list(train_label2cnt.keys()),
                y=list(train_label2cnt.values()),
                name="Train",
                marker_color="blue",
                text=list(train_label2cnt.values()),  # Add text for Train bars
                textposition='auto'  # Position text automatically (usually above the bar)
            )
        )

        fig.add_trace(
            go.Bar(
                x=list(val_label2cnt.keys()),
                y=list(val_label2cnt.values()),
                name="Val",
                marker_color="red",
                text=list(val_label2cnt.values()),  # Add text for Val bars
                textposition='auto'  # Position text automatically
            )
        )
        fig.write_image(osp.join(output_dir, "split.png"), engine="auto")

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=list(label2cnt.keys()),
                y=list(label2cnt.values()),
                name="Total",
                marker_color="green",
                text=list(label2cnt.values()),  # Add text for Total bars
                textposition='auto'  # Position text automatically
            )
        )
        fig.write_image(osp.join(output_dir, "total.png"), engine="auto")

                
if __name__ == '__main__':
    input_dir = '/DeepLearning/_athena_tests/datasets/rectangle2/raw_dataset'
    output_dir = '/DeepLearning/_athena_tests/datasets/rectangle2/split_dataset'
    image_formats = ['bmp']
    ratio = 0.3
    figs = True
    
    split_dataset(input_dir, output_dir, image_formats, ratio, figs=True)
        
            
            
            
            
            
            
            
            
            
            
        
        
        


import os
import os.path as osp
import torch 
import glob 
from PIL import Image 
import torchvision
from torchvision.transforms import functional as F, InterpolationMode

from visionsuite.engines.segmentation.utils.registry import DATASETS
from visionsuite.engines.classification.src.datasets.base_dataset import BaseDataset
import visionsuite.engines.utils.torch_utils.presets as presets

def get_transform(is_train, args):
    if is_train:
        return presets.TrainTransform(resize=args['resize'], normalize=args['normalize'], backend=args['backend'], use_v2=args['use_v2'])
    elif args['weights'] and args['test_only']:
        weights = torchvision.models.get_weight(args['weights'])
        trans = weights.transforms()

        def preprocessing(img, target):
            img = trans(img)
            size = F.get_dimensions(img)[1:]
            target = F.resize(target, size, interpolation=InterpolationMode.NEAREST)
            return img, F.pil_to_tensor(target)

        return preprocessing
    else:
        return presets.ValTransform(resize=args['resize'], normalize=args['normalize'], backend=args['backend'], use_v2=args['use_v2'])
    


@DATASETS.register()
class MaskDatasetWrapper(BaseDataset):
    def __init__(self, name="MaskDatasetWrapper", transform=None):
        if transform is None:
            mean=(0.485, 0.456, 0.406)
            std=(0.229, 0.224, 0.225)
                
            import torchvision.transforms as transforms
            transform = transforms.Compose(
                                [transforms.ToTensor(),
                                transforms.Normalize(mean=mean, std=std)]
                        )
            
        super().__init__(name=name, transform=transform)

    def load_dataset(self, train_folder_name='train', val_folder_name='val'):
        super().load_dataset()
        
        
        self.train_dataset =  DATASETS.get(self.args['train']['type'], 
                                           case_sensitive=self.args['train']['case_sensitive']
                                )(input_dir=osp.join(self.args['input_dir'], train_folder_name), 
                                    transform=get_transform(True, 
                                                    {**self.args['train']['transform'], 
                                                     **{"weights": None, "test_only": False, "backend": 'PIL', "use_v2": False}
                                                    }
                                            ),
                                    image_formats=self.args['train']['image_formats'],
                                    mask_format=self.args['train']['mask_format']
                                )
        self.log_info(f"LOADED train_dataset: {self.args['train']['type']}", self.build.__name__, __class__.__name__)
        self.log_info(f"- input_dir: {self.args['input_dir']}", self.build.__name__, __class__.__name__)
        self.log_info(f"- number of images: {len(self.train_dataset)}", self.build.__name__, __class__.__name__)
        self.log_info(f"- transforms: TODO", self.build.__name__, __class__.__name__)
        
        self.val_dataset =  DATASETS.get(self.args['val']['type'], 
                                         case_sensitive=self.args['val']['case_sensitive']
                                )(input_dir=osp.join(self.args['input_dir'], val_folder_name), 
                                    transform=get_transform(False, 
                                                    {**self.args['val']['transform'], 
                                                    **{"weights": None, "test_only": False, "backend": 'PIL', "use_v2": False}
                                                    }
                                                ),
                                    image_formats=self.args['val']['image_formats'],
                                    mask_format=self.args['val']['mask_format']
                                )
        self.log_info(f"LOADED val_dataset: {self.args['val']['type']}", self.build.__name__, __class__.__name__)
        self.log_info(f"- input_dir: {self.args['input_dir']}", self.build.__name__, __class__.__name__)
        self.log_info(f"- number of images:: {len(self.val_dataset)}", self.build.__name__, __class__.__name__)
        self.log_info(f"- transforms: TODO", self.build.__name__, __class__.__name__)
        
        self.label2index = {index: label for index, label in enumerate(self.classes)}
        self.index2label = {label: index for index, label in enumerate(self.classes)}
        self.num_classes = len(self.classes)
        self.log_info(f"- label2index: {self.label2index}", self.load_dataset.__name__, __class__.__name__)
     
        
@DATASETS.register()
class MaskDataset(torch.utils.data.Dataset):
    def __init__(self, input_dir, transform=None, image_formats=['png', 'bmp', 'jpg'], mask_format=None, **kwargs):
        self.input_dir = input_dir
        self.transform = transform
        self.image_formats = image_formats
        self.mask_format = mask_format

        self.img_files = []
        for image_format in image_formats:
            self.img_files += glob.glob(os.path.join(self.input_dir, "images/*.{}".format(image_format)))
        print(f"  - There are {len(self.img_files)} image files")
    
    def __len__(self):
        return len(self.img_files)

    def __getitem__(self, idx): 
        img_file = self.img_files[idx]
        img_ext = osp.splitext(img_file)[-1].replace('.', '')
        fname = osp.split(osp.splitext(img_file)[0])[-1]

        if self.mask_format is None:
            mask_file = osp.join(self.input_dir, f'masks/{fname}.{img_ext}'.format(fname))
        else:
            mask_file = osp.join(self.input_dir, f'masks/{fname}.{self.mask_format}'.format(fname))
            
        assert osp.exists(mask_file), RuntimeError(f"There is no such mask image: {mask_file}")

        image = Image.open(img_file)
        mask = Image.open(mask_file)        
        
        if self.transform is not None:
            image, mask = self.transform(image, mask)

        return image, mask, fname
    
    


# import os.path as osp
# import glob 
# import os
# from PIL import Image
# import torch
# from visionsuite.engines.utils.torch_utils.transforms import Compose


# def get_mask(root, image_set, transforms, use_v2=False):

#     PATHS = {
#         "train": ("train/images"),
#         "val": ("val/images"),
#     }

#     transforms = Compose([transforms])

#     img_folder = PATHS[image_set]
#     img_folder = osp.join(root, img_folder)

#     dataset = MaskDataset(img_folder, transforms=transforms)

#     # if mode == "train": #FIXME: Need to make this option 
#     #     dataset = _coco_remove_images_without_annotations(dataset, CAT_LIST)

#     return dataset


# ##FIXME: This maskdataset is too dependent to camvid dataset..., especially total_classes w/o background label 
# class MaskDataset(torch.utils.data.Dataset):
#     def __init__(self, img_folder, transforms=None, img_exts=['png', 'bmp']):
#         self.img_folder = img_folder
#         self.transforms = transforms

#         self.img_files = []
#         for img_ext in img_exts:
#             self.img_files += glob.glob(os.path.join(self.img_folder, "*.{}".format(img_ext)))
#         print(f"  - There are {len(self.img_files)} image files")
    
#     def __len__(self):
#         return len(self.img_files)

#     def __getitem__(self, idx): 
#         img_file = self.img_files[idx]
#         fname = osp.split(osp.splitext(img_file)[0])[-1]

#         mask_file = osp.join(self.img_folder, '../masks/{}.bmp'.format(fname))
#         assert osp.exists(mask_file), RuntimeError(f"There is no such mask image: {mask_file}")

#         image = Image.open(img_file)
#         mask = Image.open(mask_file)        
        
#         if self.transforms is not None:
#             image, mask = self.transforms(image, mask)

#         return image, mask, fname


def get_background_patches(
    img_width,
    img_height,
    patch_height,
    patch_width,
    points,
    overlap_ratio,
    roi=[],
    skip_highly_overlapped_tiles=True,
):

    # TODO: Need to ignore the segment of line for a negative sample?
    if len(roi) != 0 and roi is not None:
        tl_x, tl_y, br_x, br_y = roi[0], roi[1], roi[2], roi[3]
    else:
        tl_x, tl_y, br_x, br_y = 0, 0, img_width, img_height

    dx = int((1.0 - overlap_ratio) * patch_width)
    dy = int((1.0 - overlap_ratio) * patch_height)

    background_patches_rois = []  # x1y1x2y2
    background_patches_num_data = 0
    for y0 in range(tl_y, br_y, dy):
        for x0 in range(tl_x, br_x, dx):
            if y0 + patch_height > br_y:
                if skip_highly_overlapped_tiles:
                    if (y0 + patch_height - br_y) > (0.6 * patch_height):
                        continue
                    else:
                        y = br_y - patch_height
                else:
                    y = br_y - patch_height
            else:
                y = y0

            if x0 + patch_width > br_x:
                if skip_highly_overlapped_tiles:
                    if (x0 + patch_width - br_x) > (0.6 * patch_width):
                        continue
                    else:
                        x = br_x - patch_width
                else:
                    x = br_x - patch_width
            else:
                x = x0

            xmin, xmax, ymin, ymax = (
                x,
                x + patch_width,
                y,
                y + patch_height,
            )  # patch coordinates

            is_inside = False
            if len(points) > 0:
                for points_dict in points:
                    b = points_dict["points"]
                    for xb0, yb0 in b:
                        if (
                            (xb0 >= xmin)
                            and (xb0 <= xmax)
                            and (yb0 <= ymax)
                            and (yb0 >= ymin)
                        ):
                            is_inside = True
                            break
                    if is_inside:
                        break

            if is_inside:
                continue

            background_patches_rois.append([x, y, x + patch_width, y + patch_height])
            background_patches_num_data += 1

    return background_patches_rois, background_patches_num_data



import cv2 
import numpy as np
import os.path as osp 
import torch 
import imgviz

        
def save_validation(model, device, dataset, num_classes, epoch, output_dir, denormalize=False, input_channel=3, \
                        image_channel_order='bgr', validation_image_idxes_list=[], color_map=imgviz.label_colormap(50)):
    model.eval()
    origin = 25,25
    font = cv2.FONT_HERSHEY_SIMPLEX

    # if len(validation_image_idxes_list) == 0:
    #     validation_image_idxes_list = range(0, len(dataset))

    total_idx = 1
    for batch in dataset:
        if len(batch) == 3:
            image, mask, fname = batch[0].detach(), batch[1].detach(), batch[2]
        else: 
            image, mask = batch[0].detach(), batch[1].detach()
            fname = None           
        image = image.to(device, dtype=torch.float32)
        # image = image.to(device)
        image = image.unsqueeze(0)
        preds = model(image)
        if isinstance(preds, dict):
            preds = preds['out']
        elif isinstance(preds, list):
            preds = preds[0]
        
        preds = preds[0]
        preds = torch.nn.functional.softmax(preds, dim=0)
        preds = torch.argmax(preds, dim=0)
        preds = preds.detach().float().to('cpu')
        # preds.apply_(lambda x: t2l[x])
        preds = preds.numpy()

        image = image.to('cpu')[0]
        image = image.numpy()
        image = image.transpose((1, 2, 0))
        if denormalize:
            image = denormalize(image)
        # image = image.astype(np.uint8)
        mask = color_map[mask.numpy().astype(np.uint8)].astype(np.uint8)
        if input_channel == 3:
            if image_channel_order == 'rgb':
                image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            elif image_channel_order == 'bgr':
                pass
            else:
                raise ValueError(f"There is no such image_channel_order({image_channel_order})")

            preds = color_map[preds.astype(np.uint8)].astype(np.uint8)
        elif input_channel == 1:
            preds = color_map[preds.astype(np.uint8)].astype(np.uint8)

        else:
            raise NotImplementedError(f"There is not yet training for input_channel ({input_channel})")

        preds = cv2.addWeighted(image, 0.1, preds, 0.9, 0)
        mask = cv2.addWeighted(image, 0.1, mask, 0.9, 0)

        text1 = np.zeros((50, image.shape[1], input_channel), np.uint8)
        text2 = np.zeros((50, image.shape[1], input_channel), np.uint8)
        text3 = np.zeros((50, image.shape[1], input_channel), np.uint8)
        cv2.putText(text1, "(a) original", origin, font, 0.6, (255,255,255), 1)
        cv2.putText(text2, "(b) ground truth" , origin, font, 0.6, (255,255,255), 1)
        cv2.putText(text3, "(c) predicted" , origin, font, 0.6, (255,255,255), 1)

        image = cv2.vconcat([text1, image])
        mask = cv2.vconcat([text2, mask])
        preds = cv2.vconcat([text3, preds.astype(np.uint8)])

        res = cv2.hconcat([image, mask, preds])
        cv2.imwrite(osp.join(output_dir, str(epoch) + "_" + fname + '_{}.png'.format(total_idx)), res)
        total_idx += 1



import inspect
from typing import List

import torch
from visionsuite.engines.utils.registry import SCHEDULERS, Registry


def register_torch_schedulers() -> List[str]:
    TORCH_SCHEDULERS = Registry("torch_scheduler", parent=SCHEDULERS, scope="visionsuite.engines.utils.registry.schedulers")
    for module_name in dir(torch.optim.lr_scheduler):
        if module_name.startswith("__"):
            continue
        
        _object = getattr(torch.optim.lr_scheduler, module_name)
        if inspect.isclass(_object):# and issubclass(_object, torch.optim.lr_scheduler._LRScheduler):
            TORCH_SCHEDULERS.register(module=_object)

    return TORCH_SCHEDULERS


TORCH_SCHEDULERS = register_torch_schedulers()

        


import json
import math
import os.path as osp
from pathlib import Path

import cv2
import numpy as np
from PIL import Image

SHAPE_TYPE = ['polygon', 'watershed', 'point', 'rectangle', 'image']


def get_roi_info(json_file, anns, roi_info, img_h, img_w):
    assert osp.exists(json_file), ValueError(f"There is no such json-file: {json_file}")

    if osp.exists(json_file) and 'rois' in anns and isinstance(anns['rois'], list) and len(anns['rois']) > 0:
        roi_info = anns['rois']
    else:
        if roi_info is None:
            roi_info = [[0, 0, img_w, img_h]]
        else:
            roi_info = roi_info

    return roi_info


def init_labelme_json(image_fn, imgsz_w, imgsz_h, imageData=None, imageId=None):
    labelme = {}

    labelme['version'] = "2.2.0"
    labelme['flags'] = {}
    labelme['shapes'] = []
    labelme["rois"] = []
    labelme["imageId"] = imageId
    labelme["imagePath"] = image_fn
    labelme["imageData"] = imageData
    labelme["imageHeight"] = imgsz_h
    labelme["imageWidth"] = imgsz_w

    return labelme


def add_labelme_element(labelme,
                        shape_type,
                        label,
                        points,
                        rotationrect=None,
                        group_id=None,
                        bbox=None,
                        flags={},
                        otherData=None):
    anns = {}
    anns['otherData'] = otherData
    anns["label"] = label
    anns["points"] = points
    anns["rotationrect"] = rotationrect
    anns["group_id"] = group_id
    anns["shape_type"] = shape_type
    anns["flags"] = flags
    anns['bbox'] = bbox

    labelme['shapes'].append(anns)

    return labelme


def read_labelme(json_file):

    if isinstance(json_file, str):
        assert osp.exists(json_file), f"There is no labelme json file: {json_file}"
    elif isinstance(json_file, Path):
        assert json_file.exists(), f"There is no labelme json file: {json_file}"
        json_file = str(json_file)

    with open(json_file) as f_json:
        ann = json.load(f_json)

    # version = ann['version']
    # flags = ann['flags']
    # shapes = ann['shapes']
    # imagePath = ann["imagePath"]
    # imageData = ann["imageData"]
    # imageHeight = ann["imageHeight"]
    # imageWidth = ann["imageWidth"]

    return ann


def xyxy_to_polygon(points):
    """This will change rectangle(p[x1, y1], [x2, y2]]) to polygon."""
    assert isinstance(points, list), ValueError(f"The type of points should be list, not {type(points)}")
    assert len(points) == 2, ValueError(f"The length of points({points}) should be 2, not {len(points)}")
    assert len(
        points[0]) == 2, ValueError(f"The length of one of points({points[0]}) should be 2, not {len(points[0])}")

    polygon = [points[0]]
    polygon.append([points[1][0], points[0][1]])
    polygon.append(points[1])
    polygon.append([points[0][0], points[1][1]])
    polygon.append(points[0])

    return polygon


def get_polygon_points_from_labelme_shape(shape, shape_type, include_point_positive, mode):
    """
    This will consider all points including point or line(two points) to check a negative sample for false positive
    cases.

    For negative samples,
        - if mode is 'train': include or not according to includ_point_positive
        - if mode is 'val': must include
    """

    assert shape_type in SHAPE_TYPE, ValueError(f"The shape-type({shape_type}) must be one of {SHAPE_TYPE}")

    points = shape['points']
    # TODO: Need to consider for empty json
    if len(points) == 0:  # handle exception where there is no points
        return []

    if shape_type in ['polygon', 'watershed', "point"]:
        if len(points) > 0 and len(points) <= 2:  # for a negative sample
            if mode in ['train']:
                if include_point_positive:
                    return points
                else:
                    return []
            elif mode in ['test', 'val']:
                return points
            else:
                raise RuntimeError(f"There is no such mode({mode}) for datasets")

    # TODO: How to consider circle as polygon
    # elif shape_type == 'circle':  # two points: center point and one of edge point
    #     points = [points[0]]

    elif shape_type == 'rectangle':
        points = xyxy_to_polygon(points)

    else:
        raise ValueError(f"There is no such shape-type: {shape_type}")

    return points


def get_mask_from_labels(labels, width, height, class2label, format='pil'):

    mask = np.zeros((height, width))
    for label_dict in labels:
        label = label_dict['label']
        points = label_dict['points']

        if len(points) != 0:
            arr = np.array(points, dtype=np.int32)
            cv2.fillPoly(mask, [arr], color=(class2label[label]))

    if format == 'pil':
        return Image.fromarray(mask)
    elif format == 'opencv':
        return mask


def get_mask_from_labelme(json_file, width, height, class2label, format='pil', metis=None):

    if metis is None:
        with open(json_file) as f:
            anns = json.load(f)
    else:
        anns = {"shapes": metis}
    mask = np.zeros((height, width))
    for label_idx in range(0, len(class2label.keys())):
        for shapes in anns['shapes']:
            shape_type = shapes['shape_type'].lower()
            label = shapes['label'].lower()
            if label == list(class2label.keys())[label_idx]:
                _points = shapes['points']
                if shape_type == 'circle':
                    cx, cy = _points[0][0], _points[0][1]
                    radius = int(math.sqrt((cx - _points[1][0]) ** 2 + (cy - _points[1][1]) ** 2))
                    cv2.circle(mask, (int(cx), int(cy)), int(radius), True, -1)
                elif shape_type in ['rectangle']:
                    if len(_points) == 2:
                        arr = np.array(_points, dtype=np.int32)
                    else:
                        RuntimeError(f"Rectangle labeling should have 2 points")
                    cv2.fillPoly(mask, [arr], color=(class2label[label]))
                elif shape_type in ['polygon', 'watershed']:
                    if len(_points) > 2:  # handle cases for 1 point or 2 points
                        arr = np.array(_points, dtype=np.int32)
                    else:
                        continue
                    cv2.fillPoly(mask, [arr], color=(class2label[label]))
                elif shape_type in ['point']:
                    pass
                else:
                    raise ValueError(f"There is no such shape-type: {shape_type}")

    if format == 'pil':
        return Image.fromarray(mask)
    elif format == 'opencv':
        return mask
    else:
        NotImplementedError(f'There is no such case for {format}')


def get_points_from_image(image, class_list, roi, ltp, _labelme, contour_thres):
    for idx, cls_name in enumerate(class_list):
        new_mask = (image == idx + 1).astype(np.uint8)
        contours, _ = cv2.findContours(new_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        for contour in contours:
            if len(contour) < contour_thres:
                pass
            else:
                epsilon = 0.001 * cv2.arcLength(contour, True)
                approx = cv2.approxPolyDP(contour, epsilon, True)
                for objects in [approx]:
                    polygon = []
                    xs, ys = [], []
                    for jdx in range(0, len(objects)):
                        polygon.append([
                            int(objects[jdx][0][0] + roi[0] + ltp[0]),
                            int(objects[jdx][0][1] + roi[1] + ltp[1])])
                        xs.append(int(objects[jdx][0][0] + roi[0] + ltp[0]))
                        ys.append(int(objects[jdx][0][1] + roi[1] + ltp[1]))

                    bbox = [int(np.max(xs)), int(np.max(ys)), int(np.min(xs)), int(np.min(ys))]
                    _labelme = add_labelme_element(_labelme,
                                                   shape_type="polygon",
                                                   label=cls_name,
                                                   points=polygon,
                                                   bbox=bbox)
    return _labelme


def get_labeled_gt_image(gt_img, json_file, classes, color_map, logger=None):

    assert osp.exists(json_file), ValueError(f"There is no such json-file: {json_file}")

    test_y_offset = 8
    linewidth = 1
    font = cv2.FONT_HERSHEY_SIMPLEX
    font_size = 0.5

    try:
        with open(json_file) as f:
            anns = json.load(f)
    except Exception as error:
        raise Exception(f"There is something wrong in json_file: {json_file} - {error}")

    for shapes in anns['shapes']:
        label = shapes['label'].lower()
        if label in classes or label.lower() in classes:
            shape_type = shapes['shape_type'].lower()
            if shape_type in ['polygon', 'watershed', 'point']:
                if shape_type == 'polygon' or shape_type == 'watershed':
                    points = shapes['points']
                    if len(points) <= 2:
                        continue
                elif shape_type == 'point':
                    points = shapes['points']
                    if len(points) <= 1:
                        continue
                arr = np.array(points, dtype=np.int32)
                color = ([int(c) for c in color_map[classes.index(label) + 1]])
                cv2.fillPoly(gt_img, [arr], color=color)
                cv2.rectangle(gt_img, (int(np.min(arr, axis=0)[0]), int(np.min(arr, axis=0)[1])),
                              (int(np.max(arr, axis=0)[0]), int(np.max(arr, axis=0)[1])), color, linewidth)
                if np.min(arr, axis=0)[1] - 10 < 0:
                    text_y = int(np.min(arr, axis=0)[1] + 5)
                else:
                    text_y = int(np.min(arr, axis=0)[1] - 10)
                cv2.putText(gt_img, label, (int(np.min(arr, axis=0)[0]), text_y), font, font_size, color, 2)

            elif shape_type == 'rectangle':
                points = shapes['points']
                cv2.rectangle(gt_img, (int(points[0][0]), int(points[0][1])), (int(points[1][0]), int(points[1][1])),
                              color, linewidth)
                if points[0][1] - test_y_offset < 0:
                    text_y = int(points[0][1] + 5)
                else:
                    text_y = int(points[0][1] - test_y_offset)
                cv2.putText(gt_img, label, (int(points[0][0]), text_y), font, font_size, color, 2)
            elif shape_type == 'circle':
                _points = shapes['points']
                assert len(_points) == 2, 'Shape of shape_type=circle must have 2 points'
                (cx, cy), (px, py) = _points
                d = math.sqrt((cx - px) ** 2 + (cy - py) ** 2)
                cv2.circle(gt_img, (int(cx), int(cy)),
                           int(d),
                           color=([int(c) for c in color_map[classes.index(label) + 1]]),
                           thickness=-1)
                if cy - d - test_y_offset < 0:
                    text_y = int(cy - d + 5)
                else:
                    text_y = int(cy - d - test_y_offset)
                cv2.putText(gt_img, label, (int(cx - d), text_y), font, font_size, color, 2)
            else:
                if logger is not None:
                    logger.raise_error_log(ValueError, f"There is no such shape-type: {shape_type}")
                else:
                    raise ValueError(f"There is no such shape-type: {shape_type}")

    return gt_img, anns


if __name__ == '__main__':

    # json_file = "/HDD/datasets/projects/samkee/test_90_movingshot/split_dataset/val/20230213_64_Side64_94.json"
    # width = 1920
    # height = 1080
    # class2label = {'bubble': 0, 'dust': 1, 'line': 2, 'crack': 3, 'black': 4, 'peeling': 5, 'burr': 6}

    # mask = get_mask_from_labelme(json_file, width, height, class2label, 'opencv')
    # print(mask.shape)
    # cv2.imwrite("/projects/mask.png", mask)

    json_file = "/HDD/datasets/_unittests/multiple_rois/wo_patches/sungwoo_edge/split_datasets/val/122111520173660_7_EdgeDown.json"
    width = 9344
    height = 7000
    class2label = {'_background_': 0, 'stabbed': 1, 'pop': 2}

    mask = get_mask_from_labelme(json_file, width, height, class2label, 'opencv')
    print(mask.shape)
    cv2.imwrite("/projects/mask.png", mask)


from typing import Union 

from visionsuite.cores.roboflow.utils import labelme2yolo_iseg, labelme2yolo_hbb, labelme2yolo_obb

class DatasetConverter:
    
    def __init__(self, task):
        self.__task = task
        self.input_dir = None
        self.output_dir = None 
        self.copy_image = True 
        self.image_ext = 'bmp'

        self._set_converter()
        
    @property 
    def task(self):
        return self.__task 
    
    @task.setter
    def task(self, value):
        self.__task = value
    
    def _set_converter(self):
        if self.__task == 'hbb_detection':
            self._converter = labelme2yolo_hbb
        elif self.__task == 'instance_segmentation':
            self._converter = labelme2yolo_iseg
        elif self.__task == 'obb_detection':
            self._converter = labelme2yolo_obb
        else:
            raise NotImplementedError(f"There is no such converter for {self.__task}")
        
        print(f"SELECTED converter is {self._converter.__name__}")
        
    def run(self, input_dir: str, copy_image: bool=True, image_ext: Union[list, str]='bmp', output_dir: str=None):
                      
        def _set_output_dir(input_dir, output_dir):
            if output_dir is None:
                if self.__task == 'hbb_detection':
                    output_dir = input_dir + '_yolo_hbb'
                elif self.__task == 'instance_segmentation':
                    output_dir = input_dir + '_yolo_iseg'
                elif self.__task == 'obb_detection':
                    output_dir = input_dir + '_yolo_obb'

            return output_dir
        
        output_dir = _set_output_dir(input_dir, output_dir)
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.copy_image = copy_image
        self.image_ext = image_ext
        
        self._converter(input_dir=input_dir, output_dir=output_dir, copy_image=copy_image, image_ext=image_ext)
            
        
if __name__ == '__main__':
    # input_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/split_dataset'

    # copy_image = True
    # image_ext = 'bmp'
       
    # converter = DatasetConverter('obb_detection')
    # converter.run(input_dir, copy_image=copy_image, image_ext=image_ext)
    
    input_dir = '/HDD/datasets/projects/Tenneco/Metalbearing/warp/TOP_BOT_CHAMFER_WARP_ALL/241204/split_labelme_patch_dataset'

    copy_image = True
    image_ext = 'bmp'
       
    converter = DatasetConverter('hbb_detection')
    converter.run(input_dir, copy_image=copy_image, image_ext=image_ext)
    
            



from __future__ import absolute_import

import warnings

depreciated_warning = "\n----------------------------------------\n`backbones`is depreciated, use `keras_unet_collection.base` instead.\ne.g.\nfrom keras_unet_collection import base\nbase.unet_2d_base(...);\n----------------------------------------"
warnings.warn(depreciated_warning);

from keras_unet_collection._model_unet_2d import unet_2d_base as unet_2d_backbone
from keras_unet_collection._model_unet_plus_2d import unet_plus_2d_base as unet_plus_2d_backbone
from keras_unet_collection._model_r2_unet_2d import r2_unet_2d_base as r2_unet_2d_backbone
from keras_unet_collection._model_att_unet_2d import att_unet_2d_base as att_unet_2d_backbone
from keras_unet_collection._model_resunet_a_2d import resunet_a_2d_base as resunet_a_2d_backbone
from keras_unet_collection._model_u2net_2d import u2net_2d_base as u2net_2d_backbone
from keras_unet_collection._model_unet_3plus_2d import unet_3plus_2d_base as unet_3plus_2d_backbone



import copy
import datetime
import os
import time

import torch
import torch.ao.quantization
import torch.utils.data
import torchvision
import utils
from torch import nn
from train import evaluate, load_data, train_one_epoch


def main(args):
    if args.output_dir:
        utils.mkdir(args.output_dir)

    utils.init_distributed_mode(args)
    print(args)

    if args.post_training_quantize and args.distributed:
        raise RuntimeError("Post training quantization example should not be performed on distributed mode")

    # Set backend engine to ensure that quantized model runs on the correct kernels
    if args.qbackend not in torch.backends.quantized.supported_engines:
        raise RuntimeError("Quantized backend not supported: " + str(args.qbackend))
    torch.backends.quantized.engine = args.qbackend

    device = torch.device(args.device)
    torch.backends.cudnn.benchmark = True

    # Data loading code
    print("Loading data")
    train_dir = os.path.join(args.data_path, "train")
    val_dir = os.path.join(args.data_path, "val")

    dataset, dataset_test, train_sampler, test_sampler = load_data(train_dir, val_dir, args)
    data_loader = torch.utils.data.DataLoader(
        dataset, batch_size=args.batch_size, sampler=train_sampler, num_workers=args.workers, pin_memory=True
    )

    data_loader_test = torch.utils.data.DataLoader(
        dataset_test, batch_size=args.eval_batch_size, sampler=test_sampler, num_workers=args.workers, pin_memory=True
    )

    print("Creating model", args.model)
    # when training quantized models, we always start from a pre-trained fp32 reference model
    prefix = "quantized_"
    model_name = args.model
    if not model_name.startswith(prefix):
        model_name = prefix + model_name
    model = torchvision.models.get_model(model_name, weights=args.weights, quantize=args.test_only)
    model.to(device)

    if not (args.test_only or args.post_training_quantize):
        model.fuse_model(is_qat=True)
        model.qconfig = torch.ao.quantization.get_default_qat_qconfig(args.qbackend)
        torch.ao.quantization.prepare_qat(model, inplace=True)

        if args.distributed and args.sync_bn:
            model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)

        optimizer = torch.optim.SGD(
            model.parameters(), lr=args.lr, momentum=args.momentum, weight_decay=args.weight_decay
        )

        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=args.lr_step_size, gamma=args.lr_gamma)

    criterion = nn.CrossEntropyLoss()
    model_without_ddp = model
    if args.distributed:
        model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
        model_without_ddp = model.module

    if args.resume:
        checkpoint = torch.load(args.resume, map_location="cpu", weights_only=True)
        model_without_ddp.load_state_dict(checkpoint["model"])
        optimizer.load_state_dict(checkpoint["optimizer"])
        lr_scheduler.load_state_dict(checkpoint["lr_scheduler"])
        args.start_epoch = checkpoint["epoch"] + 1

    if args.post_training_quantize:
        # perform calibration on a subset of the training dataset
        # for that, create a subset of the training dataset
        ds = torch.utils.data.Subset(dataset, indices=list(range(args.batch_size * args.num_calibration_batches)))
        data_loader_calibration = torch.utils.data.DataLoader(
            ds, batch_size=args.batch_size, shuffle=False, num_workers=args.workers, pin_memory=True
        )
        model.eval()
        model.fuse_model(is_qat=False)
        model.qconfig = torch.ao.quantization.get_default_qconfig(args.qbackend)
        torch.ao.quantization.prepare(model, inplace=True)
        # Calibrate first
        print("Calibrating")
        evaluate(model, criterion, data_loader_calibration, device=device, print_freq=1)
        torch.ao.quantization.convert(model, inplace=True)
        if args.output_dir:
            print("Saving quantized model")
            if utils.is_main_process():
                torch.save(model.state_dict(), os.path.join(args.output_dir, "quantized_post_train_model.pth"))
        print("Evaluating post-training quantized model")
        evaluate(model, criterion, data_loader_test, device=device)
        return

    if args.test_only:
        evaluate(model, criterion, data_loader_test, device=device)
        return

    model.apply(torch.ao.quantization.enable_observer)
    model.apply(torch.ao.quantization.enable_fake_quant)
    start_time = time.time()
    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            train_sampler.set_epoch(epoch)
        print("Starting training for epoch", epoch)
        train_one_epoch(model, criterion, optimizer, data_loader, device, epoch, args)
        lr_scheduler.step()
        with torch.inference_mode():
            if epoch >= args.num_observer_update_epochs:
                print("Disabling observer for subseq epochs, epoch = ", epoch)
                model.apply(torch.ao.quantization.disable_observer)
            if epoch >= args.num_batch_norm_update_epochs:
                print("Freezing BN for subseq epochs, epoch = ", epoch)
                model.apply(torch.nn.intrinsic.qat.freeze_bn_stats)
            print("Evaluate QAT model")

            evaluate(model, criterion, data_loader_test, device=device, log_suffix="QAT")
            quantized_eval_model = copy.deepcopy(model_without_ddp)
            quantized_eval_model.eval()
            quantized_eval_model.to(torch.device("cpu"))
            torch.ao.quantization.convert(quantized_eval_model, inplace=True)

            print("Evaluate Quantized model")
            evaluate(quantized_eval_model, criterion, data_loader_test, device=torch.device("cpu"))

        model.train()

        if args.output_dir:
            checkpoint = {
                "model": model_without_ddp.state_dict(),
                "eval_model": quantized_eval_model.state_dict(),
                "optimizer": optimizer.state_dict(),
                "lr_scheduler": lr_scheduler.state_dict(),
                "epoch": epoch,
                "args": args,
            }
            utils.save_on_master(checkpoint, os.path.join(args.output_dir, f"model_{epoch}.pth"))
            utils.save_on_master(checkpoint, os.path.join(args.output_dir, "checkpoint.pth"))
        print("Saving models after epoch ", epoch)

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    print(f"Training time {total_time_str}")


def get_args_parser(add_help=True):
    import argparse

    parser = argparse.ArgumentParser(description="PyTorch Quantized Classification Training", add_help=add_help)

    parser.add_argument("--data-path", default="/datasets01/imagenet_full_size/061417/", type=str, help="dataset path")
    parser.add_argument("--model", default="mobilenet_v2", type=str, help="model name")
    parser.add_argument("--qbackend", default="qnnpack", type=str, help="Quantized backend: fbgemm or qnnpack")
    parser.add_argument("--device", default="cuda", type=str, help="device (Use cuda or cpu Default: cuda)")

    parser.add_argument(
        "-b", "--batch-size", default=32, type=int, help="images per gpu, the total batch size is $NGPU x batch_size"
    )
    parser.add_argument("--eval-batch-size", default=128, type=int, help="batch size for evaluation")
    parser.add_argument("--epochs", default=90, type=int, metavar="N", help="number of total epochs to run")
    parser.add_argument(
        "--num-observer-update-epochs",
        default=4,
        type=int,
        metavar="N",
        help="number of total epochs to update observers",
    )
    parser.add_argument(
        "--num-batch-norm-update-epochs",
        default=3,
        type=int,
        metavar="N",
        help="number of total epochs to update batch norm stats",
    )
    parser.add_argument(
        "--num-calibration-batches",
        default=32,
        type=int,
        metavar="N",
        help="number of batches of training set for \
                              observer calibration ",
    )

    parser.add_argument(
        "-j", "--workers", default=16, type=int, metavar="N", help="number of data loading workers (default: 16)"
    )
    parser.add_argument("--lr", default=0.0001, type=float, help="initial learning rate")
    parser.add_argument("--momentum", default=0.9, type=float, metavar="M", help="momentum")
    parser.add_argument(
        "--wd",
        "--weight-decay",
        default=1e-4,
        type=float,
        metavar="W",
        help="weight decay (default: 1e-4)",
        dest="weight_decay",
    )
    parser.add_argument("--lr-step-size", default=30, type=int, help="decrease lr every step-size epochs")
    parser.add_argument("--lr-gamma", default=0.1, type=float, help="decrease lr by a factor of lr-gamma")
    parser.add_argument("--print-freq", default=10, type=int, help="print frequency")
    parser.add_argument("--output-dir", default=".", type=str, help="path to save outputs")
    parser.add_argument("--resume", default="", type=str, help="path of checkpoint")
    parser.add_argument("--start-epoch", default=0, type=int, metavar="N", help="start epoch")
    parser.add_argument(
        "--cache-dataset",
        dest="cache_dataset",
        help="Cache the datasets for quicker initialization. \
             It also serializes the transforms",
        action="store_true",
    )
    parser.add_argument(
        "--sync-bn",
        dest="sync_bn",
        help="Use sync batch norm",
        action="store_true",
    )
    parser.add_argument(
        "--test-only",
        dest="test_only",
        help="Only test the model",
        action="store_true",
    )
    parser.add_argument(
        "--post-training-quantize",
        dest="post_training_quantize",
        help="Post training quantize the model",
        action="store_true",
    )

    # distributed training parameters
    parser.add_argument("--world-size", default=1, type=int, help="number of distributed processes")
    parser.add_argument("--dist-url", default="env://", type=str, help="url used to set up distributed training")

    parser.add_argument(
        "--interpolation", default="bilinear", type=str, help="the interpolation method (default: bilinear)"
    )
    parser.add_argument(
        "--val-resize-size", default=256, type=int, help="the resize size used for validation (default: 256)"
    )
    parser.add_argument(
        "--val-crop-size", default=224, type=int, help="the central crop size used for validation (default: 224)"
    )
    parser.add_argument(
        "--train-crop-size", default=224, type=int, help="the random crop size used for training (default: 224)"
    )
    parser.add_argument("--clip-grad-norm", default=None, type=float, help="the maximum gradient norm (default None)")
    parser.add_argument("--weights", default=None, type=str, help="the weights enum name to load")

    parser.add_argument("--backend", default="PIL", type=str.lower, help="PIL or tensor - case insensitive")
    parser.add_argument("--use-v2", action="store_true", help="Use V2 transforms")

    return parser


if __name__ == "__main__":
    args = get_args_parser().parse_args()
    if args.backend in ("fbgemm", "qnnpack"):
        raise ValueError(
            "The --backend parameter has been re-purposed to specify the backend of the transforms (PIL or Tensor) "
            "instead of the quantized backend. Please use the --qbackend parameter to specify the quantized backend."
        )
    main(args)



import pynvml
import numpy as np
import time

class GPULogger:
    def __init__(self, device_ids=[0]):
        self._device_ids = device_ids
        pynvml.nvmlInit()

        self.data = {}

    def update(self):
        for device_id in self._device_ids:
            handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)
            utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
            memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            power_info = pynvml.nvmlDeviceGetPowerUsage(handle)

            if f'GPU {device_id}' not in self.data:
                self.data[f'GPU {device_id}'] = {"Util (%)": [utilization.gpu],
                                                 "Mem. (MB)": [float(round(memory_info.used / 1024**2, 2))],
                                                 "Power (W)": [float(round(power_info / 1000, 2))]
                                            }
            else:
                self.data[f'GPU {device_id}']["Util (%)"].append(utilization.gpu),
                self.data[f'GPU {device_id}']["Mem. (MB)"].append(float(round(memory_info.used / 1024**2, 2)))
                self.data[f'GPU {device_id}']["Power (W)"].append(float(round(power_info / 1000, 2)))

    def mean(self):
        ret = {}
        for key1, val1 in self.data.items():
            ret[key1] = {}
            for key2, val2 in val1.items():
                ret[key1].update({key2: round(np.mean(val2), 3)})
            
        return ret
    
    def max(self):
        ret = {}
        for key1, val1 in self.data.items():
            ret[key1] = {}
            for key2, val2 in val1.items():
                ret[key1].update({key2: round(max(val2), 3)})
            
        return ret

    def min(self):
        ret = {}
        for key1, val1 in self.data.items():
            ret[key1] = {}
            for key2, val2 in val1.items():
                ret[key1].update({key2: round(min(val2), 3)})
            
        return ret
    
    def clear(self):
        self.data = {}

    def end(self):
        pynvml.nvmlShutdown()
        
if __name__ == '__main__':
    logger = GPULogger([0, 1])
    
    for _ in range(3600):
        logger.update()
        time.sleep(1)
        
        print(logger.max())




from visionsuite.engines.classification.utils.augment.presets import get_module
from .random_cut_mix import RandomCutMix
from .random_mix_up import RandomMixUp


def get_mixup_cutmix(*, mixup_alpha, cutmix_alpha, num_classes, use_v2):
    transforms_module = get_module(use_v2)

    mixup_cutmix = []
    if mixup_alpha > 0:
        mixup_cutmix.append(
            transforms_module.MixUp(alpha=mixup_alpha, num_classes=num_classes)
            if use_v2
            else RandomMixUp(num_classes=num_classes, p=1.0, alpha=mixup_alpha)
        )
    if cutmix_alpha > 0:
        mixup_cutmix.append(
            transforms_module.CutMix(alpha=cutmix_alpha, num_classes=num_classes)
            if use_v2
            else RandomCutMix(num_classes=num_classes, p=1.0, alpha=cutmix_alpha)
        )
    if not mixup_cutmix:
        return None

    return transforms_module.RandomChoice(mixup_cutmix)



import torch
from torchvision.transforms.functional import InterpolationMode


def get_module(use_v2):
    # We need a protected import to avoid the V2 warning in case just V1 is used
    if use_v2:
        import torchvision.transforms.v2

        return torchvision.transforms.v2
    else:
        import torchvision.transforms

        return torchvision.transforms


class ClassificationPresetTrain:
    # Note: this transform assumes that the input to forward() are always PIL
    # images, regardless of the backend parameter. We may change that in the
    # future though, if we change the output type from the dataset.
    def __init__(
        self,
        *,
        crop_size,
        mean=(0.485, 0.456, 0.406),
        std=(0.229, 0.224, 0.225),
        interpolation=InterpolationMode.BILINEAR,
        hflip_prob=0.5,
        auto_augment_policy=None,
        ra_magnitude=9,
        augmix_severity=3,
        random_erase_prob=0.0,
        backend="pil",
        use_v2=False,
    ):
        T = get_module(use_v2)

        transforms = []
        backend = backend.lower()
        if backend == "tensor":
            transforms.append(T.PILToTensor())
        elif backend != "pil":
            raise ValueError(f"backend can be 'tensor' or 'pil', but got {backend}")

        transforms.append(T.RandomResizedCrop(crop_size, interpolation=interpolation, antialias=True))
        if hflip_prob > 0:
            transforms.append(T.RandomHorizontalFlip(hflip_prob))
        if auto_augment_policy is not None:
            if auto_augment_policy == "ra":
                transforms.append(T.RandAugment(interpolation=interpolation, magnitude=ra_magnitude))
            elif auto_augment_policy == "ta_wide":
                transforms.append(T.TrivialAugmentWide(interpolation=interpolation))
            elif auto_augment_policy == "augmix":
                transforms.append(T.AugMix(interpolation=interpolation, severity=augmix_severity))
            else:
                aa_policy = T.AutoAugmentPolicy(auto_augment_policy)
                transforms.append(T.AutoAugment(policy=aa_policy, interpolation=interpolation))

        if backend == "pil":
            transforms.append(T.PILToTensor())

        transforms.extend(
            [
                T.ToDtype(torch.float, scale=True) if use_v2 else T.ConvertImageDtype(torch.float),
                T.Normalize(mean=mean, std=std),
            ]
        )
        if random_erase_prob > 0:
            transforms.append(T.RandomErasing(p=random_erase_prob))

        if use_v2:
            transforms.append(T.ToPureTensor())

        self.transforms = T.Compose(transforms)

    def __call__(self, img):
        return self.transforms(img)


class ClassificationPresetEval:
    def __init__(
        self,
        *,
        crop_size,
        resize_size=256,
        mean=(0.485, 0.456, 0.406),
        std=(0.229, 0.224, 0.225),
        interpolation=InterpolationMode.BILINEAR,
        backend="pil",
        use_v2=False,
    ):
        T = get_module(use_v2)
        transforms = []
        backend = backend.lower()
        if backend == "tensor":
            transforms.append(T.PILToTensor())
        elif backend != "pil":
            raise ValueError(f"backend can be 'tensor' or 'pil', but got {backend}")

        transforms += [
            T.Resize(resize_size, interpolation=interpolation, antialias=True),
            T.CenterCrop(crop_size),
        ]

        if backend == "pil":
            transforms.append(T.PILToTensor())

        transforms += [
            T.ToDtype(torch.float, scale=True) if use_v2 else T.ConvertImageDtype(torch.float),
            T.Normalize(mean=mean, std=std),
        ]

        if use_v2:
            transforms.append(T.ToPureTensor())

        self.transforms = T.Compose(transforms)

    def __call__(self, img):
        return self.transforms(img)



from visionsuite.engines.utils.bases.base_results import BaseResults, update_log


class Results(BaseResults):
    __accuracy = -1
    __time_for_a_epoch = -1
    __time_for_epochs = -1
    __classes_accuracy = {}

    @classmethod
    def __update_classes_accuracy(cls, value):
        assert isinstance(value, dict), ValueError(
            f"Argugments to update log must be dictionary, not{type(value)}"
        )
        cls.__classes_accuracy.update(value)

    @classmethod
    def get_attribute_names(cls):
        attribute_names = list(cls.__dict__.keys())
        return attribute_names

    def __init__(self, classes=None, logger=None):
        super().__init__()
        if classes is not None:
            for class_ in classes:
                self.__classes_accuracy[class_] = -1

            if logger is not None:
                logger.info(
                    f"The accuracy for {classes} has been set",
                    self.__ini__.__name__,
                    __class__.__name__,
                )

    @property
    def accuracy(self):
        return self.__accuracy

    @accuracy.setter
    @update_log
    def accuracy(self, value):
        assert value >= 0, ValueError(f"Train accuracy cannot be lower than 0")
        self.__accuracy = value

    @property
    def time_for_a_epoch(self):
        return self.__time_for_a_epoch

    @time_for_a_epoch.setter
    @update_log
    def time_for_a_epoch(self, value):
        self.__time_for_a_epoch = value

    @property
    def time_for_epochs(self):
        return self.__time_for_epochs

    @time_for_epochs.setter
    @update_log
    def time_for_epochs(self, value):
        self.__time_for_epochs = value

    @property
    def classes_accuracy(self):
        return self.__classes_accuracy

    @classes_accuracy.setter
    def classes_accuracy(self, value1, value2=None):
        if value2 is None:
            assert isinstance(value1, dict), ValueError(
                f"Argument for classes_accuracy must be dictionary, not {type(value1)}"
            )
            value = value1
        else:
            assert isinstance(value1, str) and isinstance(
                value2, (int, float)
            ), ValueError(f"Accuracy value must be int or float, not {type(value2)}")
            value = {value1: value2}

        self.__update_classes_accuracy(value)
        self._update_log(value)





import torchvision
from visionsuite.engines.classification.utils.registry import DATASETS
from visionsuite.engines.classification.src.datasets.base_dataset import BaseDataset


@DATASETS.register()
class CifarDataset(BaseDataset):
    def __init__(self, num_classes=10, transform=None):
        if transform is None:
            mean = (0.5, 0.5, 0.5)
            std = (0.5, 0.5, 0.5)
                
            import torchvision.transforms as transforms
            transform = transforms.Compose(
                                            [transforms.ToTensor(),
                                            transforms.Normalize(mean, std)])
        super().__init__(transform=transform)
        self.num_classes = num_classes
        
    def load_dataset(self):
        super().load_dataset()
        
        self.train_dataset = getattr(torchvision.datasets, f"CIFAR{self.num_classes}")(root='/tmp/data', 
                                                        train=True, download=True, transform=self._transform)
            
        self.val_dataset = getattr(torchvision.datasets, f"CIFAR{self.num_classes}")(root='/tmp/data', 
                                                        train=False, download=True, transform=self._transform)

        self.label2index = {index: label for index, label in enumerate(self.train_dataset.classes)}
        self.index2label = {label: index for index, label in enumerate(self.train_dataset.classes)}
        self.classes = self.train_dataset.classes
        print(f"label2index: {self.label2index}")





import torch
import torch.nn as nn
import torch.nn.functional as F


from .delta2theta import delta2theta


class LayerReg(nn.Module):
    """x has shape (B, N, C)

               B, N, C
    reshape -> B, C, A, A
    conv3x3 -> B, C, (A - 2), (A - 2)
    conv3x3 -> B, C, (A - 4), (A - 4)
    conv3x3 -> B, C, (A - 6), (A - 6)
    reshape -> B, C
    fc      -> B, out_channels
    """
    def __init__(self, in_channels=256, out_channels=2, num_convs=0, feat_size=7):
        super().__init__()
        self.num_convs = num_convs
        self.feat_size = feat_size

        if self.num_convs > 0:
            self.norms = nn.ModuleList()
            self.convs = nn.ModuleList()
            self.relu = nn.ReLU(True)
            for i in range(self.num_convs):
                self.norms.append(nn.LayerNorm([in_channels, feat_size - i * 2, feat_size - i * 2]))
                self.convs.append(nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=0))
            self.last_feat_area = (feat_size - num_convs * 2) ** 2

        self.norm_reg = nn.LayerNorm(in_channels)
        self.fc_reg = nn.Linear(in_channels, out_channels)

    def forward(self, x):

        if self.num_convs > 0:
            B, N, C = x.shape
            x = x.transpose(-2, -1).reshape(B, C, self.feat_size, self.feat_size)
            for i in range(self.num_convs):
                norm = self.norms[i]
                conv = self.convs[i]
                x = self.relu(conv(norm(x)))
            x = x.reshape(B, C, self.last_feat_area).transpose(-2, -1)

        bbox_pr = self.fc_reg(self.norm_reg(x.mean(dim=1)))
        return bbox_pr


# ===== ===== ===== ===== ===== ===== ===== ===== ===== =====


def drop_path(x, drop_prob: float = 0., training: bool = False):
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


#class Attention(nn.Module):
#    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
#        super().__init__()
#        self.num_heads = num_heads
#        head_dim = dim // num_heads
#        self.scale = qk_scale or head_dim ** -0.5

#        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
#        self.attn_drop = nn.Dropout(attn_drop)
#        self.proj = nn.Linear(dim, dim)
#        self.proj_drop = nn.Dropout(proj_drop)

#    def forward(self, x):
#        B, N, C = x.shape
#        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
#        q, k, v = qkv[0], qkv[1], qkv[2]

#        attn = ((self.scale * q) @ k.transpose(-2, -1))
#        attn = attn.softmax(dim=-1)
#        attn = self.attn_drop(attn)

#        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
#        x = self.proj(x)
#        x = self.proj_drop(x)
#        return x, attn


#class Block(nn.Module):
#    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., 
#                 attn_drop=0., drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, init_values=0):
#        super().__init__()
#        self.norm1 = norm_layer(dim)
#        self.attn = Attention(
#            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
#        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
#        self.norm2 = norm_layer(dim)
#        mlp_hidden_dim = int(dim * mlp_ratio)
#        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

#        if init_values > 0:
#            self.gamma_1 = nn.Parameter(init_values * torch.ones((dim)), requires_grad=True)
#            self.gamma_2 = nn.Parameter(init_values * torch.ones((dim)), requires_grad=True)
#        else:
#            self.gamma_1, self.gamma_2 = None, None

#    def forward(self, x, return_attention=False):
#        y, attn = self.attn(self.norm1(x))
#        if return_attention:
#            return attn
#        if self.gamma_1 is None:
#            x = x + self.drop_path(y)
#            x = x + self.drop_path(self.mlp(self.norm2(x)))
#        else:
#            x = x + self.drop_path(self.gamma_1 * y)
#            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
#        return x


# ===== ===== ===== ===== ===== ===== ===== ===== ===== =====


class AttentionSTDC(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0., am_mode_str=''):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.am_mode_str = am_mode_str

    def forward(self, x, actv_mask):
        B, N, C = x.shape

        if 'X' in self.am_mode_str:
            x = x.reshape(B, N, self.num_heads, C // self.num_heads).transpose(1, 2)
            x = x * actv_mask
            x = x.transpose(1, 2).reshape(B, N, C)

        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        if 'Q' in self.am_mode_str:
            q = q * actv_mask
        if 'K' in self.am_mode_str:
            k = k * actv_mask
        if 'V' in self.am_mode_str:
            v = v * actv_mask

        attn = ((self.scale * q) @ k.transpose(-2, -1))
        attn = attn.float().clamp(min=torch.finfo(torch.float32).min, max=torch.finfo(torch.float32).max) # new!
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn


class BlockSTDC(nn.Module):
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., 
                 attn_drop=0., drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, init_values=0, 
                 dc_mode_str        = '', 
                 have_predicted_str = '', 
                 num_convs          = 0, 
                 am_mode_str        = '', 
                 rois_mode          = 'hbbox',
                 ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = AttentionSTDC(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop, am_mode_str=am_mode_str)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if init_values > 0:
            self.gamma_1 = nn.Parameter(init_values * torch.ones((dim)), requires_grad=True)
            self.gamma_2 = nn.Parameter(init_values * torch.ones((dim)), requires_grad=True)
        else:
            self.gamma_1, self.gamma_2 = None, None

        self.num_heads = num_heads
        self.dc_mode_str = dc_mode_str
        self.have_predicted_str = have_predicted_str
        # self.num_convs = num_convs
        self.am_mode_str = am_mode_str
        self.rois_mode = rois_mode

        lack_xy = ('XY' not in self.have_predicted_str)
        lack_wh = ('WH' not in self.have_predicted_str)
        lack_a  = ('A'  not in self.have_predicted_str)
        lack_xy_o_wh_o_a = (lack_xy or lack_wh or lack_a)
        self.need_layer_reg = ((len(self.dc_mode_str) != 0) or (len(self.am_mode_str) != 0 and lack_xy_o_wh_o_a))

        if self.need_layer_reg:
            self.layer_reg = LayerReg(in_channels=dim, out_channels=5, num_convs=num_convs)
        else:
            pass

    def forward(self, x, rois, bbox_pr_xy, bbox_pr_wh, bbox_pr_a, return_attention=False):
        B, N, C = x.shape
        H = self.num_heads
        I = C // self.num_heads
        A = int(N ** 0.5)

        # (1) predict bbox
        if self.need_layer_reg:
            bbox_reg = self.layer_reg(x)
            bbox_reg_xy = bbox_reg[:, 0:2]
            bbox_reg_wh = bbox_reg[:, 2:4]
            bbox_reg_a  = bbox_reg[:, 4:5]
        else:
            pass

        # (2) return bbox_reg
        if len(self.dc_mode_str) != 0:
            bbox_pr = bbox_reg
        else:
            bbox_pr = torch.zeros([B, 0], dtype=x.dtype, device=x.device, requires_grad=x.requires_grad)

        # (3) generate actv_mask
        if len(self.am_mode_str) != 0:
            x_ones = torch.ones([B * H, I, A, A], dtype=x.dtype, device=x.device, requires_grad=False)

            if 'XY' in self.dc_mode_str:
                delta_xy = bbox_reg_xy.clone().detach()
            elif 'XY' in self.have_predicted_str:
                delta_xy = bbox_pr_xy.clone().detach()
            else:
                delta_xy = bbox_reg_xy

            if 'WH' in self.dc_mode_str:
                delta_wh = bbox_reg_wh.clone().detach()
            elif 'WH' in self.have_predicted_str:
                delta_wh = bbox_pr_wh.clone().detach()
            else:
                delta_wh = bbox_reg_wh

            if 'A' in self.dc_mode_str:
                delta_a = bbox_reg_a.clone().detach()
            elif 'A' in self.have_predicted_str:
                delta_a = bbox_pr_a.clone().detach()
            else:
                delta_a = bbox_reg_a

            deltas = torch.cat([delta_xy, delta_wh, delta_a], dim=1)

            theta_c = delta2theta(rois=rois[:, 1:], deltas=deltas, rois_mode=self.rois_mode)
            theta_c = theta_c.unsqueeze(1).expand(-1, H, -1).reshape(B * H, 2, 3)

            grid = F.affine_grid(theta_c, x_ones.size())
            grid = grid.type(x_ones.type())                                 # avoid fp16/fp32 confusion
            actv_mask = F.grid_sample(x_ones, grid) # + x_ones * 0.1        # B * H, I, A, A

            actv_mask = actv_mask.reshape(B, H, I, N).transpose(-2, -1)     # B, H, N, I
        else:
            actv_mask = None

        y, attn = self.attn(self.norm1(x), actv_mask)
        if return_attention:
            return attn
        if self.gamma_1 is None:
            x = x + self.drop_path(y)
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        else:
            x = x + self.drop_path(self.gamma_1 * y)
            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
        return x, bbox_pr




# from mmrotate/core/visualization/image.py


# Copyright (c) OpenMMLab. All rights reserved.

import cv2
import matplotlib.pyplot as plt
import mmcv
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from mmdet.core.visualization import palette_val
from mmdet.core.visualization.image import draw_labels, draw_masks

from mmrotate.core.visualization.palette import get_palette

EPS = 1e-2


def _get_adaptive_scales(areas, min_area=800, max_area=30000):
    """Get adaptive scales according to areas.

    The scale range is [0.5, 1.0]. When the area is less than
    ``'min_area'``, the scale is 0.5 while the area is larger than
    ``'max_area'``, the scale is 1.0.

    Args:
        areas (ndarray): The areas of bboxes or masks with the
            shape of (n, ).
        min_area (int): Lower bound areas for adaptive scales.
            Default: 800.
        max_area (int): Upper bound areas for adaptive scales.
            Default: 30000.

    Returns:
        ndarray: The adaotive scales with the shape of (n, ).
    """
    scales = 0.5 + (areas - min_area) / (max_area - min_area)
    scales = np.clip(scales, 0.5, 1.0)
    return scales


def draw_rbboxes(ax, bboxes, color='g', alpha=0.8, thickness=2):
    """Draw oriented bounding boxes on the axes.

    Args:
        ax (matplotlib.Axes): The input axes.
        bboxes (ndarray): The input bounding boxes with the shape
            of (n, 5).
        color (list[tuple] | matplotlib.color): the colors for each
            bounding boxes.
        alpha (float): Transparency of bounding boxes. Default: 0.8.
        thickness (int): Thickness of lines. Default: 2.

    Returns:
        matplotlib.Axes: The result axes.
    """
    polygons = []
    for i, bbox in enumerate(bboxes):
        xc, yc, w, h, ag = bbox[:5]
        wx, wy = w / 2 * np.cos(ag), w / 2 * np.sin(ag)
        hx, hy = -h / 2 * np.sin(ag), h / 2 * np.cos(ag)
        p1 = (xc - wx - hx, yc - wy - hy)
        p2 = (xc + wx - hx, yc + wy - hy)
        p3 = (xc + wx + hx, yc + wy + hy)
        p4 = (xc - wx + hx, yc - wy + hy)
        poly = np.int0(np.array([p1, p2, p3, p4]))
        polygons.append(Polygon(poly))
    p = PatchCollection(
        polygons,
        facecolor='none',
        edgecolors=color,
        linewidths=thickness,
        alpha=alpha)
    ax.add_collection(p)

    return ax


def imshow_det_rbboxes(img,
                       bboxes=None,
                       labels=None,
                       segms=None,
                       class_names=None,
                       score_thr=0,
                       bbox_color='green',
                       text_color='green',
                       mask_color=None,
                       thickness=2,
                       font_size=13,
                       win_name='',
                       show=True,
                       wait_time=0,
                       out_file=None):
    """Draw bboxes and class labels (with scores) on an image.

    Args:
        img (str | ndarray): The image to be displayed.
        bboxes (ndarray): Bounding boxes (with scores), shaped (n, 5) or
            (n, 6).
        labels (ndarray): Labels of bboxes.
        segms (ndarray | None): Masks, shaped (n,h,w) or None.
        class_names (list[str]): Names of each classes.
        score_thr (float): Minimum score of bboxes to be shown. Default: 0.
        bbox_color (list[tuple] | tuple | str | None): Colors of bbox lines.
           If a single color is given, it will be applied to all classes.
           The tuple of color should be in RGB order. Default: 'green'.
        text_color (list[tuple] | tuple | str | None): Colors of texts.
           If a single color is given, it will be applied to all classes.
           The tuple of color should be in RGB order. Default: 'green'.
        mask_color (list[tuple] | tuple | str | None, optional): Colors of
           masks. If a single color is given, it will be applied to all
           classes. The tuple of color should be in RGB order.
           Default: None.
        thickness (int): Thickness of lines. Default: 2.
        font_size (int): Font size of texts. Default: 13.
        show (bool): Whether to show the image. Default: True.
        win_name (str): The window name. Default: ''.
        wait_time (float): Value of waitKey param. Default: 0.
        out_file (str, optional): The filename to write the image.
            Default: None.

    Returns:
        ndarray: The image with bboxes drawn on it.
    """
    assert bboxes is None or bboxes.ndim == 2, \
        f' bboxes ndim should be 2, but its ndim is {bboxes.ndim}.'
    assert labels.ndim == 1, \
        f' labels ndim should be 1, but its ndim is {labels.ndim}.'
    assert bboxes is None or bboxes.shape[1] == 5 or bboxes.shape[1] == 6, \
        f' bboxes.shape[1] should be 5 or 6, but its {bboxes.shape[1]}.'
    assert bboxes is None or bboxes.shape[0] <= labels.shape[0], \
        'labels.shape[0] should not be less than bboxes.shape[0].'
    assert segms is None or segms.shape[0] == labels.shape[0], \
        'segms.shape[0] and labels.shape[0] should have the same length.'
    assert segms is not None or bboxes is not None, \
        'segms and bboxes should not be None at the same time.'

    img = mmcv.imread(img).astype(np.uint8)

    if score_thr > 0:
        assert bboxes is not None and bboxes.shape[1] == 6
        scores = bboxes[:, -1]
        inds = scores > score_thr
        bboxes = bboxes[inds, :]
        labels = labels[inds]
        if segms is not None:
            segms = segms[inds, ...]

    img = mmcv.bgr2rgb(img)
    width, height = img.shape[1], img.shape[0]
    img = np.ascontiguousarray(img)

    fig = plt.figure(win_name, frameon=False)
    plt.title(win_name)
    canvas = fig.canvas
    dpi = fig.get_dpi()
    # add a small EPS to avoid precision lost due to matplotlib's truncation
    # (https://github.com/matplotlib/matplotlib/issues/15363)
    fig.set_size_inches((width + EPS) / dpi, (height + EPS) / dpi)

    # remove white edges by set subplot margin
    plt.subplots_adjust(left=0, right=1, bottom=0, top=1)
    ax = plt.gca()
    ax.axis('off')

    max_label = int(max(labels) if len(labels) > 0 else 0)
    text_palette = palette_val(get_palette(text_color, max_label + 1))
    text_colors = [text_palette[label] for label in labels]

    num_bboxes = 0
    if bboxes is not None:
        num_bboxes = bboxes.shape[0]
        bbox_palette = palette_val(get_palette(bbox_color, max_label + 1))
        colors = [bbox_palette[label] for label in labels[:num_bboxes]]
        draw_rbboxes(ax, bboxes, colors, alpha=0.8, thickness=thickness)

        #horizontal_alignment = 'left'
        #positions = bboxes[:, :2].astype(np.int32) + thickness
        #areas = bboxes[:, 2] * bboxes[:, 3]
        #scales = _get_adaptive_scales(areas)
        #scores = bboxes[:, 5] if bboxes.shape[1] == 6 else None
        #draw_labels(
        #    ax,
        #    labels[:num_bboxes],
        #    positions,
        #    scores=scores,
        #    class_names=class_names,
        #    color=text_colors,
        #    font_size=font_size,
        #    scales=scales,
        #    horizontal_alignment=horizontal_alignment)

    if segms is not None:
        mask_palette = get_palette(mask_color, max_label + 1)
        colors = [mask_palette[label] for label in labels]
        colors = np.array(colors, dtype=np.uint8)
        draw_masks(ax, img, segms, colors, with_edge=True)

        if num_bboxes < segms.shape[0]:
            segms = segms[num_bboxes:]
            horizontal_alignment = 'center'
            areas = []
            positions = []
            for mask in segms:
                _, _, stats, centroids = cv2.connectedComponentsWithStats(
                    mask.astype(np.uint8), connectivity=8)
                largest_id = np.argmax(stats[1:, -1]) + 1
                positions.append(centroids[largest_id])
                areas.append(stats[largest_id, -1])
            areas = np.stack(areas, axis=0)
            scales = _get_adaptive_scales(areas)
            draw_labels(
                ax,
                labels[num_bboxes:],
                positions,
                class_names=class_names,
                color=text_colors,
                font_size=font_size,
                scales=scales,
                horizontal_alignment=horizontal_alignment)

    plt.imshow(img)

    stream, _ = canvas.print_to_buffer()
    buffer = np.frombuffer(stream, dtype='uint8')
    img_rgba = buffer.reshape(height, width, 4)
    rgb, alpha = np.split(img_rgba, [3], axis=2)
    img = rgb.astype('uint8')
    img = mmcv.rgb2bgr(img)

    if show:
        # We do not use cv2 for display because in some cases, opencv will
        # conflict with Qt, it will output a warning: Current thread
        # is not the object's thread. You can refer to
        # https://github.com/opencv/opencv-python/issues/46 for details
        if wait_time == 0:
            plt.show()
        else:
            plt.show(block=False)
            plt.pause(wait_time)
    if out_file is not None:
        mmcv.imwrite(img, out_file)

    plt.close()

    return img



# -*- coding: utf-8 -*-
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from .layers import unetConv2, unetUp_origin
from .init_weights import init_weights
from torchvision import models


class UNet2Plus(nn.Module):
    def __init__(self, n_channels=3, n_classes=1, bilinear=True, feature_scale=4, 
                 is_deconv=True, is_batchnorm=True, is_ds=True):
        super(UNet_2Plus, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear
        self.feature_scale = feature_scale
        self.is_deconv = is_deconv
        self.is_batchnorm = is_batchnorm
        self.is_ds = is_ds
        filters = [64, 128, 256, 512, 1024]

        # downsampling
        self.conv00 = unetConv2(self.n_channels, filters[0], self.is_batchnorm)
        self.maxpool0 = nn.MaxPool2d(kernel_size=2)
        self.conv10 = unetConv2(filters[0], filters[1], self.is_batchnorm)
        self.maxpool1 = nn.MaxPool2d(kernel_size=2)
        self.conv20 = unetConv2(filters[1], filters[2], self.is_batchnorm)
        self.maxpool2 = nn.MaxPool2d(kernel_size=2)
        self.conv30 = unetConv2(filters[2], filters[3], self.is_batchnorm)
        self.maxpool3 = nn.MaxPool2d(kernel_size=2)
        self.conv40 = unetConv2(filters[3], filters[4], self.is_batchnorm)

        # upsampling
        self.up_concat01 = unetUp_origin(filters[1], filters[0], self.is_deconv)
        self.up_concat11 = unetUp_origin(filters[2], filters[1], self.is_deconv)
        self.up_concat21 = unetUp_origin(filters[3], filters[2], self.is_deconv)
        self.up_concat31 = unetUp_origin(filters[4], filters[3], self.is_deconv)

        self.up_concat02 = unetUp_origin(filters[1], filters[0], self.is_deconv, 3)
        self.up_concat12 = unetUp_origin(filters[2], filters[1], self.is_deconv, 3)
        self.up_concat22 = unetUp_origin(filters[3], filters[2], self.is_deconv, 3)

        self.up_concat03 = unetUp_origin(filters[1], filters[0], self.is_deconv, 4)
        self.up_concat13 = unetUp_origin(filters[2], filters[1], self.is_deconv, 4)

        self.up_concat04 = unetUp_origin(filters[1], filters[0], self.is_deconv, 5)

        # final conv (without any concat)
        self.final_1 = nn.Conv2d(filters[0], n_classes, 1)
        self.final_2 = nn.Conv2d(filters[0], n_classes, 1)
        self.final_3 = nn.Conv2d(filters[0], n_classes, 1)
        self.final_4 = nn.Conv2d(filters[0], n_classes, 1)

        # initialise weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init_weights(m, init_type='kaiming')
            elif isinstance(m, nn.BatchNorm2d):
                init_weights(m, init_type='kaiming')


    def forward(self, inputs):
        # column : 0
        X_00 = self.conv00(inputs)
        maxpool0 = self.maxpool0(X_00)
        X_10 = self.conv10(maxpool0)
        maxpool1 = self.maxpool1(X_10)
        X_20 = self.conv20(maxpool1)
        maxpool2 = self.maxpool2(X_20)
        X_30 = self.conv30(maxpool2)
        maxpool3 = self.maxpool3(X_30)
        X_40 = self.conv40(maxpool3)

        # column : 1
        X_01 = self.up_concat01(X_10, X_00)
        X_11 = self.up_concat11(X_20, X_10)
        X_21 = self.up_concat21(X_30, X_20)
        X_31 = self.up_concat31(X_40, X_30)

        # column : 2
        X_02 = self.up_concat02(X_11, X_00, X_01)
        X_12 = self.up_concat12(X_21, X_10, X_11)
        X_22 = self.up_concat22(X_31, X_20, X_21)

        # column : 3
        X_03 = self.up_concat03(X_12, X_00, X_01, X_02)
        X_13 = self.up_concat13(X_22, X_10, X_11, X_12)

        # column : 4
        X_04 = self.up_concat04(X_13, X_00, X_01, X_02, X_03)

        # final layer
        final_1 = self.final_1(X_01)
        final_2 = self.final_2(X_02)
        final_3 = self.final_3(X_03)
        final_4 = self.final_4(X_04)
        final = (final_1 + final_2 + final_3 + final_4) / 4

        if self.is_ds:
            return F.sigmoid(final)
        else:
            return F.sigmoid(final_4)


if __name__ == '__main__':
    model = UNet2Plus()
    print('# generator parameters:', 1.0 * sum(param.numel() for param in model.parameters()) / 1000000)
    
    params = list(model.named_parameters())
    for i in range(len(params)):
        name, param = params[i]
        print('name:', name, ' param.shape:', param.shape)
